//
// Copyright (c) 2011-2012, Matthijs van Leeuwen
//
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
// Neither the name of the Universiteit Utrecht, KU Leuven, nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
#include "../../elobal.h"

#if defined (__unix__)
#include <glibc_s.h>
#endif

#include <StringUtils.h>

#include "../dataset/DataSet.h"
#include "../datasubset/DataSubset.h"
#include "../data/BinaryColumn.h"
#include "../data/FloatColumn.h"
#include "../data/NominalColumn.h"
#include "../subgroup/Subgroup.h"

#include "ConditionArray.h"

bool *ConditionArray::AllTrue = NULL;

ConditionArray::ConditionArray() {
	mSize = 0;
	mColumnIds = NULL;
	mColumnTypes = NULL;
	mTypes = NULL;
	mValues = NULL;
	mParent = NULL;
}
ConditionArray::ConditionArray(const uint32 size) {
	mSize = size;
	mColumnIds = new uint32[mSize];
	mColumnTypes = new ColumnType[mSize];
	mTypes = new ConditionType[mSize];
	mValues = new ConditionValue[mSize];
	mParent = NULL;
}
ConditionArray::ConditionArray(ConditionArray *ca) {
	mSize = ca->mSize;
	mColumnIds = new uint32[mSize];
	memcpy(mColumnIds, ca->mColumnIds, mSize * sizeof(uint32));
	mColumnTypes = new ColumnType[mSize];
	memcpy(mColumnTypes, ca->mColumnTypes, mSize * sizeof(ColumnType));
	mTypes = new ConditionType[mSize];
	memcpy(mTypes, ca->mTypes, mSize * sizeof(ConditionType));
	mValues = new ConditionValue[mSize];
	memcpy(mValues, ca->mValues, mSize * sizeof(ConditionValue));
	mParent = ca->GetParent();
}
ConditionArray::ConditionArray(Subgroup *subgroup, const uint32 columnId, const ConditionType t, const bool val) {
	ConditionArray *ca = subgroup->GetConditions();
	mSize = ca->mSize;
	mColumnIds = new uint32[mSize+1];
	memcpy(mColumnIds, ca->mColumnIds, mSize * sizeof(uint32));
	mColumnTypes = new ColumnType[mSize+1];
	memcpy(mColumnTypes, ca->mColumnTypes, mSize * sizeof(ColumnType));
	mTypes = new ConditionType[mSize+1];
	memcpy(mTypes, ca->mTypes, mSize * sizeof(ConditionType));
	mValues = new ConditionValue[mSize+1];
	memcpy(mValues, ca->mValues, mSize * sizeof(ConditionValue));

	mColumnIds[mSize] = columnId;
	mColumnTypes[mSize] = BinaryColumnType;
	mTypes[mSize] = t;
	mValues[mSize].binVal = val;
	++mSize;

	mParent = subgroup;
}
ConditionArray::ConditionArray(Subgroup *subgroup, const uint32 columnId, const ConditionType t, const float val) {
	ConditionArray *ca = subgroup->GetConditions();
	mSize = ca->mSize;
	mColumnIds = new uint32[mSize+1];
	memcpy(mColumnIds, ca->mColumnIds, mSize * sizeof(uint32));
	mColumnTypes = new ColumnType[mSize+1];
	memcpy(mColumnTypes, ca->mColumnTypes, mSize * sizeof(ColumnType));
	mTypes = new ConditionType[mSize+1];
	memcpy(mTypes, ca->mTypes, mSize * sizeof(ConditionType));
	mValues = new ConditionValue[mSize+1];
	memcpy(mValues, ca->mValues, mSize * sizeof(ConditionValue));

	mColumnIds[mSize] = columnId;
	mColumnTypes[mSize] = FloatColumnType;
	mTypes[mSize] = t;
	mValues[mSize].floatVal = val;
	++mSize;

	mParent = subgroup;
}
ConditionArray::ConditionArray(Subgroup *subgroup, const uint32 columnId, const ConditionType t, const uint32 val) {
	ConditionArray *ca = subgroup->GetConditions();
	mSize = ca->mSize;
	mColumnIds = new uint32[mSize+1];
	memcpy(mColumnIds, ca->mColumnIds, mSize * sizeof(uint32));
	mColumnTypes = new ColumnType[mSize+1];
	memcpy(mColumnTypes, ca->mColumnTypes, mSize * sizeof(ColumnType));
	mTypes = new ConditionType[mSize+1];
	memcpy(mTypes, ca->mTypes, mSize * sizeof(ConditionType));
	mValues = new ConditionValue[mSize+1];
	memcpy(mValues, ca->mValues, mSize * sizeof(ConditionValue));

	mColumnIds[mSize] = columnId;
	mColumnTypes[mSize] = NominalColumnType;
	mTypes[mSize] = t;
	mValues[mSize].nomVal = val;
	++mSize;

	mParent = subgroup;
}
ConditionArray::ConditionArray(ConditionArray *ca, const uint32 condId) {
	mSize = ca->mSize - 1;
	mColumnIds = new uint32[mSize];
	mColumnTypes = new ColumnType[mSize];
	mTypes = new ConditionType[mSize];
	mValues = new ConditionValue[mSize];
	uint32 idx = 0;
	for(uint32 i=0; i<=mSize; i++) {
		if(i==condId)
			continue;
		mColumnIds[idx] = ca->mColumnIds[i];
		mColumnTypes[idx] = ca->mColumnTypes[i];
		mTypes[idx] = ca->mTypes[i];
		mValues[idx] = ca->mValues[i];
		++idx;
	}
}
ConditionArray::~ConditionArray() {
	if(mSize > 0) {
		delete[] mColumnIds;
		delete[] mColumnTypes;
		delete[] mTypes;
		delete[] mValues;
	}
}
bool ConditionArray::IsColumnUsed(const uint32 colId) {
	for(uint32 i=0; i<mSize; i++)
		if(colId == mColumnIds[i])
			return true;
	return false;
}
bool ConditionArray::ContainsEqualsForColumn(const uint32 colId) {
	for(uint32 i=0; i<mSize; i++)
		if(colId == mColumnIds[i] && mTypes[i] == Equal)
			return true;
	return false;
}
bool ConditionArray::Contains(const uint32 colId, const ConditionType t, const bool val) {
	for(uint32 i=0; i<mSize; i++)
		if(colId == mColumnIds[i] && mTypes[i] == t && mValues[i].binVal == val)
			return true;
	return false;
}
bool ConditionArray::Contains(const uint32 colId, const ConditionType t, const float val) {
	for(uint32 i=0; i<mSize; i++)
		if(colId == mColumnIds[i] && mTypes[i] == t && mValues[i].floatVal == val)
			return true;
	return false;
}
bool ConditionArray::Contains(const uint32 colId, const ConditionType t, const uint32 val) {
	for(uint32 i=0; i<mSize; i++)
		if(colId == mColumnIds[i] && mTypes[i] == t && mValues[i].nomVal == val)
			return true;
	return false;
}
DataSubset *ConditionArray::Apply(DataSet *ds, bool *baseMask /* = NULL */) { // When mask is specified, assume that only the *last* condition needs to be applied!!
	uint32 firstCondition;
	uint32 numRows = ds->GetNumRows();
	bool *mask = new bool[numRows];

	if(baseMask == NULL) {
		// Start with all-true mask
		memcpy(mask, AllTrue, numRows*sizeof(bool));
		firstCondition = 0;
	} else {
		memcpy(mask, baseMask, numRows*sizeof(bool));
		firstCondition = mSize - 1; // do only latest added condition
	}

	// Get column info
	Id2IdxMap id2idx = ds->GetColumnId2IdxMap();
	Column **columns = ds->GetColumns();
	Column *column;

	// Foreach condition
	for(uint32 i=firstCondition; i<mSize; i++) {
		column = columns[id2idx[mColumnIds[i]]];

		switch(mColumnTypes[i]) {
			case BinaryColumnType:
				ApplyBinary((BinaryColumn *)column, mask, mTypes[i], mValues[i].binVal);
				break;
			case FloatColumnType:
				ApplyFloat((FloatColumn *)column, mask, mTypes[i], mValues[i].floatVal);
				break;
			case NominalColumnType:
				ApplyNominal((NominalColumn *)column, mask, mTypes[i], mValues[i].nomVal);
				break;
			default:
				THROW("Unsupported column type?!");
		}
	}

	// Return subset
	return DataSubset::Create(ds, mask, true);
}	
void ConditionArray::ApplyBinary(BinaryColumn *column, bool *mask, const ConditionType type, const bool val) {
	uint32 numRows = column->GetSize();
	bool *values = column->GetBoolValues();
	switch(type) {
		case Equal:
			for(uint32 i=0; i<numRows; i++)
				mask[i] = mask[i] && (val==values[i]);
			break;
		default:
			THROW("This condition type shouldn't be used with Binary.");
			break;
	}
}
void ConditionArray::ApplyFloat(FloatColumn *column, bool *mask, const ConditionType type, const float val) {
	uint32 numRows = column->GetSize();
	float *values = column->GetFloatValues();
	switch(type) {
		case LowerThan:
			for(uint32 i=0; i<numRows; i++)
				mask[i] = mask[i] && (values[i] < val);
			break;
		case GreaterThan:
			for(uint32 i=0; i<numRows; i++)
				mask[i] = mask[i] && (values[i] > val);
			break;
		default:
			THROW("This condition type shouldn't be used with Float.");
			break;
	}
}
void ConditionArray::ApplyNominal(NominalColumn *column, bool *mask, const ConditionType type, const uint32 val) {
	uint32 numRows = column->GetSize();
	uint32 *values = column->GetNominalValues();
	switch(type) {
		case Equal:
			for(uint32 i=0; i<numRows; i++)
				mask[i] = mask[i] && (val==values[i]);
			break;
		case NotEqual:
			for(uint32 i=0; i<numRows; i++)
				mask[i] = mask[i] && (val!=values[i]);
			break;
		default:
			THROW("This condition type shouldn't be used with Nominal.");
			break;
	}
}
bool ConditionArray::Equals(ConditionArray *ca) {
	// Sizes match?
	if(mSize != ca->mSize)
		return false;

	// Try to find each column (may only occur once, currently)
	for(uint32 i=0; i<mSize; i++) {
		bool found = false;
		for(uint32 j=0; j<mSize; j++) {
			if(mColumnIds[i] == ca->mColumnIds[j]) {
				if(mTypes[i] != ca->mTypes[j])
					return false;
				if(mValues[i].nomVal != ca->mValues[j].nomVal)
					return false;
				found = true;
				break;
			}
		}
		if(!found)
			return false;
	}
	return true;
}
uint32 ConditionArray::Difference(ConditionArray *ca) {
	// returns the number of conditions that is different 
	// (if unequally sized: counts the number of conditions in the largest that is not in the smallest)
	ConditionArray *ca1, *ca2;
	uint32 diff = 0;

	// Sizes?
	if(mSize >= ca->mSize) {
		ca1 = this;
		ca2 = ca;
	} else {
		ca1 = ca;
		ca2 = this;
	}

	// Try to find each column (may only occur once, currently)
	for(uint32 i=0; i<ca1->mSize; i++) {
		bool found = false;
		for(uint32 j=0; j<ca2->mSize; j++) {
			if(ca1->mColumnIds[i] == ca2->mColumnIds[j] && ca1->mTypes[i] == ca2->mTypes[j] && ca1->mValues[i].nomVal == ca2->mValues[j].nomVal) {
				found = true;
				break;
			}
		}
		if(!found)
			diff++;
	}
	return diff;
}
string ConditionArray::ToString(DataSet *ds) {
	string s = ""; char temp[100];
	Id2IdxMap id2idx = ds->GetColumnId2IdxMap();
	Column **columns = ds->GetColumns();
	Column *column;

	for(uint32 i=0; i<mSize; i++) {
		if(i>0)
			s += " && ";

		column = columns[id2idx[mColumnIds[i]]];
		s += column->GetLabel();

		switch(mTypes[i]) {
			case Equal:
				s += " = ";
				break;
			case NotEqual:
				s += " != ";
				break;
			case LowerThan:
				s += " < ";
				break;
			case GreaterThan:
				s += " > ";
				break;
			default:
				break;
		}

		switch(mColumnTypes[i]) {
			case BinaryColumnType:
				 s += (mValues[i].binVal ? ((BinaryColumn *)column)->GetYesDesc() : ((BinaryColumn *)column)->GetNoDesc());
				break;
			case FloatColumnType:
				sprintf_s(temp, 100, "%.05f", mValues[i].floatVal);
				s += temp;
				break;
			case NominalColumnType:
				s += ((NominalColumn *)column)->GetValueDesc(mValues[i].nomVal);
				break;
			default:
				THROW("Unsupported column type?!");
		}
	}
	return s;
}
ConditionArray *ConditionArray::FromString(const string &description, DataSet *ds) {
	// Obtain tokens
	uint32 numTokens;
	string *tokens = StringUtils::Tokenize(description, numTokens);

	// Check whether a meaningful number of tokens was found
	uint32 numConditions = (numTokens + 1) / 4;
	if(numConditions * 4 - 1 != numTokens) {
		printf("Parse error: incorrect number of tokens.\n");
		return NULL;
	}

	// Init array
	ConditionArray *cond = new ConditionArray(numConditions);

	// Attempt to parse each condition
	bool error = false;
	for(uint32 i=0; i<numConditions; i++)
		if(!ParseCondition(cond, i, tokens[i*4], tokens[i*4+1], tokens[i*4+2], ds))
			error = true;

	delete[] tokens;
	if(error) {
		delete cond;
		return NULL;
	}
	return cond;
}
bool ConditionArray::ParseCondition(ConditionArray *cond, const uint32 idx, const string &label, const string &type, string &val, DataSet *ds) {
	// Identify the column
	uint32 numColumns = ds->GetNumColumns();
	Column **cols = ds->GetColumns();
	Column *column = NULL;
	for(uint32 c=0; c<numColumns; c++)
		if(cols[c]->GetLabel().compare(label)==0) {
			column = cols[c];
			break;
		}
	if(column == NULL) {
		printf("Column label could not be found: %s\n", label.c_str());
		return false;
	}

	// Fill column info
	cond->mColumnIds[idx] = column->GetId();
	cond->mColumnTypes[idx] = column->GetType();

	// Now identify the type
	if(type.compare("=")==0)
		cond->mTypes[idx] = Equal;
	else if(type.compare("!=")==0)
		cond->mTypes[idx] = NotEqual;
	else if(type.compare("<")==0)
		cond->mTypes[idx] = LowerThan;
	else if(type.compare(">")==0)
		cond->mTypes[idx] = GreaterThan;
	else {
		printf("Unsupported condition type: %s\n", type.c_str());
		return false;
	}

	// Proceed based on column type
	switch(column->GetType()) {
		case BinaryColumnType:
			if(cond->mTypes[idx] != Equal) {
				printf("Binary columns currently only support =, not %s\n", type.c_str());
				return false;
			}
			cond->mValues[idx].binVal = ((BinaryColumn*)column)->DescToValue(val);;
			break;
		case FloatColumnType:
			if(cond->mTypes[idx] != LowerThan && cond->mTypes[idx] != GreaterThan) {
				printf("Floating columns currently only support < and >, not %s\n", type.c_str());
				return false;
			}
			cond->mValues[idx].floatVal = (float) strtod(val.c_str(), NULL);
			break;
		case NominalColumnType:
			if(cond->mTypes[idx] != Equal && cond->mTypes[idx] != NotEqual) {
				printf("Nominal columns currently only support = and !=, not %s\n", type.c_str());
				return false;
			}
			cond->mValues[idx].nomVal = ((NominalColumn*)column)->DescToValue(val);
			if(cond->mValues[idx].nomVal == UINT32_MAX_VALUE) {
				printf("Nominal value could not be parsed: %s\n", val.c_str());
				return false;
			}
			break;
		default:
			THROW("Unsupported column type?!");
	}

	return true;
}

void ConditionArray::Init(const uint32 maxSize) {
	AllTrue = new bool[maxSize];
	for(uint32 i=0; i<maxSize; i++)
		AllTrue[i] = true;
}
void ConditionArray::Cleanup() {
	delete[] AllTrue;
}
