//
// Copyright (c) 2011-2012, Matthijs van Leeuwen
//
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
// Neither the name of the Universiteit Utrecht, KU Leuven, nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
#ifndef __CONDITIONARRAY_H
#define __CONDITIONARRAY_H

#include <Primitives.h>
#include <Bass.h>

enum ConditionType {
	Equal,
	NotEqual,
	LowerThan,
	GreaterThan,
};
enum ColumnType;
union ConditionValue {
	bool binVal;
	uint32 nomVal;
	float floatVal;
};

class DataSet;
class DataSubset;
class BinaryColumn;
class FloatColumn;
class NominalColumn;
class Subgroup;

class ConditionArray {
public:
	ConditionArray();							// empty array
	ConditionArray(const uint32 size);			// array of given size
	ConditionArray(ConditionArray *ca);			// copy
	ConditionArray(Subgroup *subgroup, const uint32 columnId, const ConditionType t, const bool val); // copy + add one binary condition
	ConditionArray(Subgroup *subgroup, const uint32 columnId, const ConditionType t, const float val); // copy + add one float condition
	ConditionArray(Subgroup *subgroup, const uint32 columnId, const ConditionType t, const uint32 val); // copy + add one nominal condition
	ConditionArray(ConditionArray *ca, const uint32 condId);								// copy - remove one condition
	virtual ~ConditionArray();
	ConditionArray *Clone()		{ return new ConditionArray(this); }

	// Basic info
	uint32		GetSize()		{ return mSize; }
	Subgroup	*GetParent()	{ return mParent; }
	uint32		GetColumnId(const uint32 cond) { return mColumnIds[cond]; }
	uint32		GetLastColumnId() { return mColumnIds[mSize-1]; }

	// Normal usage
	bool		IsColumnUsed(const uint32 colId);
	bool		ContainsEqualsForColumn(const uint32 colId);
	bool		Contains(const uint32 colId, const ConditionType t, const bool val);
	bool		Contains(const uint32 colId, const ConditionType t, const float val);
	bool		Contains(const uint32 colId, const ConditionType t, const uint32 val);

	DataSubset	*Apply(DataSet *ds, bool *baseMask = NULL); // when baseMask is specified, only the *last* condition is applied!

	bool		Equals(ConditionArray *ca);
	uint32		Difference(ConditionArray *ca); // returns the number of conditions that is different (if unequally sized: counts the number of conditions in the largest that is not in the smallest)

	// To/from string
	string		ToString(DataSet *ds);
	static ConditionArray *FromString(const string &description, DataSet *ds);

	// Static
	static void		Init(const uint32 maxSize);
	static void		Cleanup();

protected:
	void			ApplyBinary(BinaryColumn *column, bool *mask, const ConditionType type, const bool val);
	void			ApplyFloat(FloatColumn *column, bool *mask, const ConditionType type, const float val);
	void			ApplyNominal(NominalColumn *column, bool *mask, const ConditionType type, const uint32 val);

	uint32			mSize;

	uint32			*mColumnIds;
	ColumnType		*mColumnTypes;
	ConditionType	*mTypes;
	ConditionValue	*mValues;

	Subgroup		*mParent;

	// Static
	static bool		ParseCondition(ConditionArray *cond, const uint32 idx, const string &label, const string &type, string &val, DataSet *ds);
	static bool		*AllTrue;
};

#endif // __CONDITIONARRAY_H
