//
// Copyright (c) 2011-2012, Matthijs van Leeuwen
//
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
// Neither the name of the Universiteit Utrecht, KU Leuven, nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
#include "../../elobal.h"

#if defined (__unix__)
#include <glibc_s.h>
#endif

#include <StringUtils.h>
#include <Bass.h>
#include <logger/Log.h>

#include "../data/BinaryColumn.h"
#include "../data/FloatColumn.h"
#include "../data/NominalColumn.h"
#include "../data/Row.h"

#include "DataSet.h"

#include "ArffFile.h"

ArffFile::ArffFile() {
	mCachedDataSet = NULL;
	mCachedHeader = "";
	mCachedRows = NULL;
}

ArffFile::~ArffFile() {
	if(mCachedDataSet != NULL)
		CleanupCache();
}
DataSet* ArffFile::Read(string filename) {
	ECHO(2, printf("** Dataset :: \n"));
	uint32 pos = (uint32) filename.find_last_of("\\/");
	ECHO(2, printf(" * File:\t\t%s\n", pos==string::npos ? filename.substr(filename.length()-55,filename.length()-1).c_str() : filename.substr(pos+1,filename.length()-1).c_str()));
	FILE *fp = NULL;
	char *bp = NULL, *buffer = new char[ARFFFILE_BUFFER_LENGTH + 1];
	if(fopen_s(&fp, filename.c_str(), "r") != 0)
		THROW("Could not open file for reading.");

	DataSet *ds = NULL;

	if(fp != NULL) {
		ds = new DataSet();

		// Path & filename
		string path, ext;
		DataSet::DeterminePathFileExt(path, filename, ext);
		ds->SetPath(path);
		ds->SetFilename(filename);

		ECHO(3, printf(" * Obtaining metadata:\tin progress..."));

		// Read header
		uint32 numColumns = 0;
		vector<string> attributeDefs;
		while(!feof(fp)) {
			fgets(buffer, ARFFFILE_BUFFER_LENGTH, fp);
			StringUtils::TrimNewline(buffer);
			string s = StringUtils::Trim(buffer, " ");
			if(s.size() > 0 && s[0] == '@') {
				if(s.compare("@data") == 0 || s.compare("@DATA") == 0) // end of header
					break; 
				string declaration = s.substr(1, 8);
				StringUtils::ToLowerCase(declaration);
				if(declaration.compare("relation") == 0)
					; // skip relation name
				else if(declaration.compare("attribut") == 0) {
					attributeDefs.push_back(s);
					++numColumns;
				}
			}
		}

		// Count number of rows
		uint32 numRows = 0;
		while(!feof(fp)) {
			if(fgets(buffer, ARFFFILE_BUFFER_LENGTH, fp) == NULL)
				break;
			StringUtils::TrimNewline(buffer);
			string s = StringUtils::Trim(buffer, " ");
			if(s.length() > 0)
				++numRows;
		}

		// Go back to start of data (ftell and fseek does not work in current setup, at least not with UNIX text files)
		fseek(fp, 0, SEEK_SET);
		while(!feof(fp)) {
			fgets(buffer, ARFFFILE_BUFFER_LENGTH, fp);
			StringUtils::TrimNewline(buffer);
			string s = StringUtils::Trim(buffer, " ");
			if(s.compare("@data") == 0 || s.compare("@DATA") == 0) // end of header
				break; 
		}


		ECHO(3, printf("\r * Obtaining metadata:\tdone.         \n"));

		ECHO(3, printf(" * Initialising data structures:\tin progress..."));

		uint32 numTokens;
		uint32 numColBinary = 0, numColFloat = 0, numColNominal = 0;
		Column **columns = new Column *[numColumns];
		ColumnType *colTypes = new ColumnType[numColumns];
		for(uint32 c=0; c<numColumns; c++) {
			string* tokens = StringUtils::Tokenize(attributeDefs[c], numTokens, " ", " ");
			if(numTokens != 3) THROW("Problem in attribute metadata: "+ attributeDefs[c]);

			if(tokens[2].compare("string") == 0) {				// String
				THROW("String datatype not yet supported.");

			} else if(tokens[2].compare("integer") == 0) {		// Integer
				LOG_WARNING("Integers treated as continuous numeric data (i.e. floats).");
				columns[c] = Column::Create(FloatColumnType, c, numRows, tokens[1]);
				colTypes[c] = FloatColumnType;
				++numColFloat;

			} else if(tokens[2].compare("numeric") == 0) {		// Numeric
				columns[c] = Column::Create(FloatColumnType, c, numRows, tokens[1]);
				colTypes[c] = FloatColumnType;
				++numColFloat;

			} else {									// Nominal / boolean
				string s = tokens[2].substr(1, tokens[2].length() - 2); // remove starting { and trailing }
				uint32 numValues;
				string *valueDescs = StringUtils::Tokenize(s, numValues, ",", " ");
				if(numValues == 2) { // *always* treat as binary!
					BinaryColumn *col = (BinaryColumn *)Column::Create(BinaryColumnType, c, numRows, tokens[1]);
					if(valueDescs[0].compare("0")==0 || valueDescs[1].compare("0")==0)
						col->SetValueDescs("0", "1");
					else if(valueDescs[0].compare("false")==0 || valueDescs[1].compare("false")==0)
						col->SetValueDescs("false", "true");
					else if(valueDescs[0].compare("no")==0 || valueDescs[1].compare("no")==0)
						col->SetValueDescs("no", "yes");
					else
						col->SetValueDescs(valueDescs[0], valueDescs[1]); // no boolean values detected, but treat as binary nevertheless
					columns[c] = col;
					colTypes[c] = BinaryColumnType;
					++numColBinary;
				} else {					// Nominal
					NominalColumn *col = (NominalColumn *)Column::Create(NominalColumnType, c, numRows, tokens[1]);
					col->SetValueDescs(valueDescs, numValues);
					columns[c] = col;
					colTypes[c] = NominalColumnType;
					++numColNominal;
				}
				delete[] valueDescs;
			}
			delete[] tokens;
		}

		Row **rows = new Row *[numRows];
		for(uint32 r=0; r<numRows; r++)
			rows[r] = new Row(numColumns, r); // r becomes tid

		ECHO(3, printf("\r * Initialising data structures:\tdone          \n"));
		ECHO(2, printf(" * Database:\t\t%dr, %dc (%db, %df, %dn)\n", numRows, numColumns, numColBinary, numColFloat, numColNominal));
		ECHO(3, printf(" * Reading data:\tin progress..."));

		float val;
		for(uint32 r=0; !feof(fp) && r<numRows; ) {
			fgets(buffer, ARFFFILE_BUFFER_LENGTH, fp);
			StringUtils::TrimNewline(buffer);
			string ln = StringUtils::Trim(buffer, " ");
			if(ln.length() > 0) {
				uint32 numVals;
				string *valueDescs = StringUtils::Tokenize(ln, numVals, ",", " ");
				if(numVals != numColumns)
					THROW("Number of values read not equal to number of columns.");
				for(uint32 c=0; c<numColumns; c++) {
					val = columns[c]->SetValueFromDesc(r, valueDescs[c]);
					rows[r]->SetValue(c, val);
				}
				delete[] valueDescs;
				++r;
			}
		}

		ECHO(3, printf("\r * Reading data:\tdone          \n"));

		ds->SetColumns(columns, colTypes, numColumns);
		ds->SetRows(rows, numRows);

	} else {	// File open error!
		delete[] buffer;
		THROW("Could not read database: " + filename);
	}

	delete[] buffer;

	ECHO(3, printf("** Finished reading database.\n"));
	ECHO(2, printf("\n"));
	return ds;
}

void ArffFile::CacheDataSetForWrite(DataSet *ds) {
	mCachedDataSet = ds;
	mCachedHeader = BuildHeader(ds);

	uint32 numRows = ds->GetNumRows();
	Row **rows = ds->GetRows();
	Column **cols = ds->GetColumns();
	mCachedRows = new string[numRows];
	for(uint32 r=0; r<numRows; r++)
		mCachedRows[r] = RowToString(rows[r], cols);
}
void ArffFile::CleanupCache() {
	delete[] mCachedRows;
	mCachedDataSet = NULL;
}
void ArffFile::CachedWriteExtraColumn(const string &dsName, Column *extraColumn, string path /* = "" */) {
	if(mCachedDataSet == NULL)
		THROW("No dataset cached.");
	if(path.length() == 0)
		path = mCachedDataSet->GetPath();
	
	string fullPath = path + dsName + "." + DataFile::TypeToExt(DataFileTypeArff);
	string s;

	ECHO(3, printf("** Writing ARFF dataset to file\n"));
	size_t pos = fullPath.find_last_of("\\/");
	// Print max 55 characters of `fullpath`, if that long and includes a directory-separator cut off at that spot.
	ECHO(3, printf(" * File:\t\t%s\n", (fullPath.length() > 55 ? (pos == string::npos ? fullPath.substr(fullPath.length()-55,fullPath.length()-1)  : fullPath.substr(pos+1,fullPath.length()-1)) :	fullPath).c_str() ));

	FILE *fp;
	if(fopen_s(&fp, fullPath.c_str(), "w") != 0)
		THROW("Could not write dataset.");

	// cached header
	fwrite(mCachedHeader.c_str(), 1, mCachedHeader.length(), fp);

	// extra column
	s = GetAttributeDescription(extraColumn);
	fwrite(s.c_str(), 1, s.length(), fp);

	// start data
	fwrite("\n@data\n", 1, 7, fp);
	for(uint32 r=0; r<mCachedDataSet->GetNumRows(); r++) {
		fwrite(mCachedRows[r].c_str(), 1, mCachedRows[r].length(), fp);

		// add last extra column
		s = "," + extraColumn->GetValueString(r);
		fwrite(s.c_str(), 1, s.length(), fp);
		
		fwrite("\n", 1, 1, fp);
	}

	fclose(fp);

	ECHO(3, printf("** Finished writing dataset.\n\n"));
}
void ArffFile::CachedWriteExtraBool(const string &dsName, bool *boolValues, string path /* = "" */) {
	if(mCachedDataSet == NULL)
		THROW("No dataset cached.");
	if(path.length() == 0)
		path = mCachedDataSet->GetPath();

	string fullPath = path + dsName + "." + DataFile::TypeToExt(DataFileTypeArff);
	string s;

	ECHO(3, printf("** Writing ARFF dataset to file\n"));
	size_t pos = fullPath.find_last_of("\\/");
	// Print max 55 characters of `fullpath`, if that long and includes a directory-separator cut off at that spot.
	ECHO(3, printf(" * File:\t\t%s\n", (fullPath.length() > 55 ? (pos == string::npos ? fullPath.substr(fullPath.length()-55,fullPath.length()-1)  : fullPath.substr(pos+1,fullPath.length()-1)) :	fullPath).c_str() ));

	FILE *fp;
	if(fopen_s(&fp, fullPath.c_str(), "w") != 0)
		THROW("Could not write dataset.");

	// cached header
	fwrite(mCachedHeader.c_str(), 1, mCachedHeader.length(), fp);

	// extra column
	s = "@attribute group {0,1}\n";
	fwrite(s.c_str(), 1, s.length(), fp);

	// start data
	fwrite("\n@data\n", 1, 7, fp);
	for(uint32 r=0; r<mCachedDataSet->GetNumRows(); r++) {
		fwrite(mCachedRows[r].c_str(), 1, mCachedRows[r].length(), fp);

		// add last extra column
		if(boolValues[r])
			fwrite(",1\n", 1, 3, fp);
		else
			fwrite(",0\n", 1, 3, fp);
	}

	fclose(fp);

	ECHO(3, printf("** Finished writing dataset.\n\n"));
}
void ArffFile::CachedWriteDataSubset(const string &dsName, uint32 *rowIds, uint32 numRows, string path /* = "" */) {
	if(mCachedDataSet == NULL)
		THROW("No dataset cached.");
	if(path.length() == 0)
		path = mCachedDataSet->GetPath();
	string fullPath = path + dsName + "." + DataFile::TypeToExt(DataFileTypeArff);

	ECHO(2, printf("** Writing ARFF dataset to file\n"));
	size_t pos = fullPath.find_last_of("\\/");
	// Print max 55 characters of `fullpath`, if that long and includes a directory-separator cut off at that spot.
	ECHO(2, printf(" * File:\t\t%s\n", (fullPath.length() > 55 ? (pos == string::npos ? fullPath.substr(fullPath.length()-55,fullPath.length()-1)  : fullPath.substr(pos+1,fullPath.length()-1)) :	fullPath).c_str() ));

	FILE *fp;
	if(fopen_s(&fp, fullPath.c_str(), "w") != 0)
		THROW("Could not write dataset.");

	fwrite(mCachedHeader.c_str(), 1, mCachedHeader.length(), fp);
	fwrite("\n@data\n", 1, 7, fp);
	for(uint32 r=0; r<numRows; r++) {
		fwrite(mCachedRows[rowIds[r]].c_str(), 1, mCachedRows[rowIds[r]].length(), fp);
		fwrite("\n", 1, 1, fp);
	}

	fclose(fp);

	ECHO(2, printf("** Finished writing dataset subset.\n\n"));
}
string ArffFile::BuildHeader(DataSet *ds) {
	string h = "@relation " + ds->GetName() + "\n\n";

	Column **cols = ds->GetColumns();
	uint32 numCols = ds->GetNumColumns();
	for(uint32 c=0; c<numCols; c++)
		h += GetAttributeDescription(cols[c]);

	return h;
}
string ArffFile::GetAttributeDescription(Column *col) {
	string s, no = "", yes = "";
	s += "@attribute " + col->GetLabel() + " ";
	switch(col->GetType()) {
			case BinaryColumnType:
				((BinaryColumn *)col)->GetValueDescs(no, yes);
				s += "{" + no + "," + yes + "}";
				break;
			case FloatColumnType:
				s += "numeric";
				break;
			case NominalColumnType:
				uint32 numDescs;
				string *descs;
				descs = ((NominalColumn *)col)->GetValueDescs(numDescs);
				s += "{";
				for(uint32 i=0; i<numDescs; i++) {
					if(i > 0) s += ",";
					s += descs[i];
				}
				s += "}";
				break;
	}
	s += "\n";
	return s;
}

string ArffFile::RowToString(Row *r, Column **cols) {
	char temp[100];
	uint32 numCols;
	float *vals = r->GetValues(numCols);
	string s = "";

	for(uint32 i=0; i<numCols; i++) {
		if(i > 0) s.append(",");
		switch(cols[i]->GetType()) {
			case BinaryColumnType:
				s += (vals[i] == 1.0) ? ((BinaryColumn *)cols[i])->GetYesDesc() : ((BinaryColumn *)cols[i])->GetNoDesc();
				break;
			case FloatColumnType:
				sprintf_s(temp, 100, "%f", vals[i]);
				s.append(temp);
				break;
			case NominalColumnType:
				s += ((NominalColumn *)cols[i])->GetValueDesc((uint32)vals[i]);
				break;
		}
	}
	return s;
}
