//
// Copyright (c) 2011-2012, Matthijs van Leeuwen
//
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
// Neither the name of the Universiteit Utrecht, KU Leuven, nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
#include "../../elobal.h"

#if defined (__unix__)
#include <glibc_s.h>
#endif

#include <FileUtils.h>
#include <RandomUtils.h>
#include <StringUtils.h>
#include <Bass.h>
#include <db/Database.h>
#include <itemstructs/ItemSet.h>

#include "../data/BinaryColumn.h"
#include "../data/Row.h"
#include "../datasubset/DataSubset.h"
#include "../../models/Model.h"

#include "DataSet.h"

DataSet::DataSet() {
	mRows = NULL;
	mNumRows = 0;

	mColumns = NULL;
	mColumnTypes = NULL;
	mNumColumns = 0;

	mWeights = NULL;

	mModel = NULL;

	mName = "";
	mFilename = "";
	mPath = "";
	mRowLabels = NULL;

	mItemSetCache = NULL;
}

DataSet::DataSet(DataSet *ds) {
	mNumRows = ds->GetNumRows();
	mRows = new Row *[mNumRows];
	Row **rows = ds->GetRows();
	for(uint32 r=0; r<mNumRows; r++)
		mRows[r] = rows[r]->Clone();

	mNumColumns = ds->GetNumColumns();
	mColumns = new Column *[mNumColumns];
	Column **cols = ds->GetColumns();
	for(uint32 c=0; c<mNumColumns; c++)
		mColumns[c] = cols[c]->Clone();
	mColumnTypes = new ColumnType[mNumColumns];
	memcpy(mColumnTypes, ds->mColumnTypes, mNumColumns * sizeof(ColumnType));

	if(ds->HasWeights()) {
		mWeights = new float[mNumRows];
		memcpy(mWeights, ds->mWeights, mNumRows * sizeof(float));
	} else
		mWeights = NULL;

	mModel = ds->GetModel()==NULL ? NULL : ds->GetModel()->Clone();

	mName = ds->GetName();
	mFilename = ds->GetFilename();
	mPath = ds->GetPath();
	if(ds->HasRowLabels()) {
		mRowLabels = new string[mNumRows];
		string *labels = ds->GetRowLabels();
		for(uint32 r=0; r<mNumRows; r++) 
			mRowLabels[r] = labels[r];
	} else {
		mRowLabels = NULL;
	}

	mItemSetCache = NULL;
}

DataSet::DataSet(const string &name, Column **cols, const uint32 numColumns) {
	if(numColumns == 0)
		THROW("Cannot build dataset with 0 columns.");

	mNumColumns = numColumns;
	mColumns = new Column *[mNumColumns];
	mColumnTypes = new ColumnType[mNumColumns];
	for(uint32 c=0; c<mNumColumns; c++) {
		mColumns[c] = cols[c]->Clone();
		mColumnTypes[c] = mColumns[c]->GetType();
	}

	mWeights = NULL;

	mModel = NULL;

	mName = name;
	mFilename = name + "-" + RandomUtils::UniformNumericalString(10);
	mPath = Bass::GetTempDir();

	// Construct rows from vertical representation (not very efficient, don't do this too often)
	mNumRows = cols[0]->GetSize();
	mRows = new Row *[mNumRows];
	for(uint32 r=0; r<mNumRows; r++) {
		mRows[r] = new Row(numColumns, r);
		for(uint32 c=0; c<mNumColumns; c++)
			mRows[r]->SetValue(c, mColumns[c]->GetValue(r));
	}
	mRowLabels = NULL;

	mItemSetCache = NULL;
}

DataSet::~DataSet() {
	for(uint32 i=0; i<mNumRows; i++)
		delete mRows[i];
	delete[] mRows;
	delete[] mRowLabels;
	for(uint32 i=0; i<mNumColumns; i++)
		delete mColumns[i];
	delete[] mColumns;
	delete[] mColumnTypes;
	if(mWeights != NULL)
		delete[] mWeights;
	delete mModel;
	if(mItemSetCache != NULL) {
		for(uint32 i=0; i<mNumRows; i++)
			delete mItemSetCache[i];
		delete[] mItemSetCache;
	}
}

DataSet* DataSet::Clone() {
	return new DataSet(this);
}
Column*	DataSet::GetColumnWithId(const uint32 id) {
	for(uint32 i=0; i<mNumColumns; i++)
		if(mColumns[i]->GetId() == id)
			return mColumns[i];
	THROW("Column with specified id not found.");
}
uint32*	DataSet::GetColumnIds() {
	uint32 *ids = new uint32[mNumColumns];
	for(uint32 i=0; i<mNumColumns; i++)
		ids[i] = mColumns[i]->GetId();
	return ids;
}
Id2IdxMap DataSet::GetColumnId2IdxMap() {
	Id2IdxMap map;
	for(uint32 i=0; i<mNumColumns; i++)
		map[mColumns[i]->GetId()] = i;
	return map;
}
bool DataSet::AllColumnsOfType(const ColumnType colType) {
	for(uint32 i=0; i<mNumColumns; i++)
		if(mColumns[i]->GetType() != colType)
			return false;
	return true;
}
void DataSet::InitWeights(const float w) {
	if(mWeights!=NULL)
		delete[] mWeights;
	mWeights = new float[mNumRows];
	for(uint32 r=0; r<mNumRows; r++)
		mWeights[r] = w;
}
void DataSet::MultiplyWeights(DataSubset *subset, const float multiplier) {
	bool *mask = subset->GetMask();
	for(uint32 r=0; r<mNumRows; r++)
		if(mask[r])
			mWeights[r] *= multiplier;
}
float DataSet::ComputeSubsetWeight(DataSubset *subset) {
	float w = 0.0f;
	bool *mask = subset->GetMask();
	for(uint32 r=0; r<mNumRows; r++)
		if(mask[r])
			w += mWeights[r];
	w /= subset->GetSize();
	return w;
}
Database* DataSet::CreateDatabase() {
	return CreateDatabase(mColumns, mNumColumns);
}
Database* DataSet::CreateDatabase(Column **columns, const uint32 numColumns) {
	// Obtain index for quicker value retrieval
	uint32 *colIdx = new uint32[numColumns];
	uint32 id;
	for(uint32 i=0; i<numColumns; i++) {
		id = columns[i]->GetId();
		for(uint32 j=0; j<mNumColumns; j++) {
			if(mColumns[j]->GetId() == id) {
				colIdx[i] = j;
				break;
			}
		}
	}

	// Determine fic datatype
	ItemSetType dataType = numColumns <= 128 ? BM128ItemSetType : Uint16ItemSetType;

	// Create itemsets
	ItemSet **iss = new ItemSet *[mNumRows];
	uint16 nCol = (uint16)numColumns;
	for(uint32 r=0; r<mNumRows; r++) {
		iss[r] = ItemSet::CreateEmpty(dataType, numColumns);
		for(uint16 c=0; c<nCol; c++) {
			if(mRows[r]->GetBinaryValue(colIdx[c]))
				iss[r]->AddItemToSet(c);
		}
	}

	// Cleanup
	delete[] colIdx;

	// Return database with custom alphabet
	Database *db = new Database(iss, mNumRows, false);
	alphabet *alph = new alphabet();
	for(uint32 v=0; v<numColumns; v++)
		alph->insert(alphabetEntry(v, 0));
	db->SetAlphabet(alph);
	db->CountAlphabet();
	db->ComputeEssentials();
	return db;
}
ItemSet* DataSet::ProvideItemset(const uint32 row) {
	if(mItemSetCache == NULL)
		CacheItemsets();

	return mItemSetCache[row];
}
void DataSet::CacheItemsets() {
	// Determine fic datatype
	ItemSetType dataType = mNumColumns <= 128 ? BM128ItemSetType : Uint16ItemSetType;

	// Create cache
	mItemSetCache = new ItemSet *[mNumRows];

	// Create itemsets
	uint16 nCol = (uint16)mNumColumns;
	for(uint32 r=0; r<mNumRows; r++) {
		mItemSetCache[r] = ItemSet::CreateEmpty(dataType, mNumColumns);
		for(uint16 c=0; c<nCol; c++) {
			if(mRows[r]->GetBinaryValue(c))
				mItemSetCache[r]->AddItemToSet(c);
		}
	}
}
void DataSet::RemoveSubset(DataSubset *subset) {
	uint32 newNumRows = mNumRows - subset->GetSize();
	BinaryColumn *rowMask = subset->GetBinaryColumn();

	// Rows
	Row **newRows = new Row *[newNumRows];
	uint32 rowIdx = 0;
	for(uint32 r=0; r<mNumRows; r++) {
		if(rowMask->GetBoolValue(r)) {	// exclude
			delete mRows[r];
		} else {						// include
			newRows[rowIdx++] = mRows[r];
		}
	}
	if(rowIdx != newNumRows)
		THROW("Something fishy going on here.");
	delete[] mRows;
	mRows = newRows;
	mNumRows = newNumRows;

	// Columns
	for(uint32 c=0; c<mNumColumns; c++)
		mColumns[c]->RemoveSubset(rowMask, subset->GetSize());

	delete rowMask;
}
void DataSet::RemoveComplement(DataSubset *subset) {
	uint32 newNumRows = subset->GetSize();
	BinaryColumn *rowMask = subset->GetBinaryColumn();

	// Rows
	Row **newRows = new Row *[newNumRows];
	uint32 rowIdx = 0;
	for(uint32 r=0; r<mNumRows; r++) {
		if(rowMask->GetBoolValue(r)) {	// include
			newRows[rowIdx++] = mRows[r];
		} else {						// exclude
			delete mRows[r];
		}
	}
	delete[] mRows;
	mRows = newRows;
	mNumRows = newNumRows;

	// Columns
	for(uint32 c=0; c<mNumColumns; c++)
		mColumns[c]->RemoveComplement(rowMask, subset->GetSize());

	delete rowMask;
}
void DataSet::SetModel(Model *model) {
	if(mModel != NULL)
		delete mModel;
	mModel = model;
}
DataSet* DataSet::ReadDataSet(const string &dsName, const string &path /* = "" */, const string &baseName /*= ""*/) {
	string inDsPath = path;
	string inDsFileName = dsName;
	string inDsFileExtension = "";

	string inDsFullPath = DeterminePathFileExt(inDsPath, inDsFileName, inDsFileExtension);

	// - Check whether designated file exists
	if(!FileUtils::FileExists(inDsFullPath))
		THROW("Dataset file does not seem to exist.");

	// - Determine type of file to load
	DataFileType inDsType = DataFile::ExtToType(inDsFileExtension);
	if(inDsType == DataFileTypeNone)
		THROW("Unknown dataset type.");

	// - Create appropriate DataFile to load that type of database
	DataFile *dataFile = DataFile::Create(inDsType);

	DataSet *ds = NULL;
	try {	// Memory-leak-protection-try/catch
		ds = dataFile->Read(inDsFullPath);
		ds->SetName(inDsFileName.substr(0,inDsFileName.find_last_of('.')));
	} catch(...) {	
		delete dataFile;	
		throw;
	}
	delete dataFile;

	if(!ds)
		THROW("Could not read dataset: " + inDsFullPath);

	// Load labels if available
	ds->LoadLabels(baseName);

	return ds;
}

void DataSet::LoadLabels(const string &altName /*= ""*/) {
	delete[] mRowLabels;
	mRowLabels = NULL;

	string labelFile = GetPath() + GetName() + ".labels";
	if(!FileUtils::FileExists(labelFile) && altName.length() > 0)
		labelFile = GetPath() + altName + ".labels";
	if(FileUtils::FileExists(labelFile)) {
		FILE *fp = NULL;
		char *bp = NULL, *buffer = new char[LABELFILE_BUFFER_LENGTH + 1];
		if(fopen_s(&fp, labelFile.c_str(), "r") != 0)
			THROW("Could not open label file for reading.");

		if(fp != NULL) {
			mRowLabels = new string[mNumRows];
			uint32 idx = 0;

			while(!feof(fp) && idx<mNumRows) {
				fgets(buffer, LABELFILE_BUFFER_LENGTH, fp);
				StringUtils::TrimNewline(buffer);
				string s = StringUtils::Trim(buffer, " ");
				if(s.size() > 0)
					mRowLabels[idx++] = s;
			}

		} else {	// File open error!
			delete[] buffer;
			THROW("Could not read labelfile: " + labelFile);
		}

		delete[] buffer;
		fclose(fp);
	}
}


// filename may or may not contain extension
string DataSet::DeterminePathFileExt(string &path, string &filename, string &extension) {
	// - Split path from filename (if required, does nothing otherwise)
	FileUtils::SplitPathFromFilename(filename, path);

	// DB files -moeten- een extensie hebben.
	if(filename.find('.') == string::npos) {
		// iig geen extensie in de (file)name
		if(extension.length() == 0) { 
			// - Geen aparte extensie meegegeven, default
			filename += ".arff";
			extension = "arff";
		} else {	
			// - Aparte extensie meegegeven, zit per definitie nog niet in filename (nl geen '.' in filename) 
			if(extension[0] == '.')
				extension = extension.substr(1);
			filename += "." + extension;
		}
	} else {
		// - Wel een punt-stuk in de (file)name, niet per definitie een extensie
		string posExt = filename.substr(filename.find_last_of('.')+1);
		if(extension.length() == 0) {	
			// - Geen aparte extensie meegegeven, wel puntStuk in filename
			if(DataFile::ExtToType(posExt) == DataFileTypeNone) {
				// - Geen bekende extensie, default
				filename += ".arff";
				extension = "arff";
			} else {
				// - Bekende extensie, zal 'm wel zijn dan
				extension = posExt;
				if(extension[0] == '.')
					extension = extension.substr(1);
			}
		} else {	
			// - Wel aparte extensie meegegeven
			if(extension[0] == '.')
				extension = extension.substr(1);
			if(posExt.compare(extension) != 0) {
				// - Alleen als anders dan opgegeven, voeg 'm maar toe
				filename += "." + extension;
			}	
		}
	}


	// - Determine `inDbFullPath`
	string fullPath = path + filename;
	return fullPath;
}
