//
// Copyright (c) 2011-2012, Matthijs van Leeuwen
//
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
// Neither the name of the Universiteit Utrecht, KU Leuven, nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
#ifndef __DATASET_H
#define __DATASET_H

#include "DataFile.h"

class Row;
class Database;
class DataSubset;
class ItemSet;
class Model;

#define LABELFILE_BUFFER_LENGTH 500

#include "../data/Column.h"

typedef std::map<uint32,uint32> Id2IdxMap;

class DataSet {
public:
	DataSet();
	DataSet(DataSet *dataset);
	DataSet(const string &name, Column **columns, const uint32 numColumns);
	virtual ~DataSet();

	virtual DataSet		*Clone();

	// -- Data manipulation --

	Database*			CreateDatabase();
	Database*			CreateDatabase(Column **columns, const uint32 numColumns);
	ItemSet*			ProvideItemset(const uint32 row);

	virtual void		RemoveSubset(DataSubset *subset);
	virtual void		RemoveComplement(DataSubset *subset);

	// -- Get/Set --
	Row**				GetRows()					{ return mRows; }
	bool				HasRowLabels()				{ return mRowLabels!=NULL; }
	string*				GetRowLabels()				{ return mRowLabels; }
	uint32				GetNumRows()				{ return mNumRows; }
	void				SetRows(Row **rs, const uint32 numRows) { mNumRows = numRows; mRows = rs; }
	Column**			GetColumns() const 			{ return mColumns; }
	uint32				GetNumColumns()				{ return mNumColumns; }
	void				SetColumns(Column **cs, ColumnType *coltypes, const uint32 numCols) { mNumColumns = numCols; mColumnTypes = coltypes; mColumns = cs; }

	Column*				GetColumnWithId(const uint32 id);
	uint32*				GetColumnIds();
	Id2IdxMap			GetColumnId2IdxMap();

	bool				AllColumnsOfType(const ColumnType colType);
	bool				AllColumnsBinary()		{ return AllColumnsOfType(BinaryColumnType); }
	bool				AllColumnsNominal()		{ return AllColumnsOfType(NominalColumnType); }
	bool				AllColumnsNumeric()		{ return AllColumnsOfType(FloatColumnType); }

	bool				HasWeights()				{ return mWeights!=NULL; }
	float*				GetWeights()				{ return mWeights; }
	void				SetWeights(float *w)		{ mWeights = w; }
	void				InitWeights(const float w);
	void				MultiplyWeights(DataSubset *subset, const float multiplier);
	float				ComputeSubsetWeight(DataSubset *subset);

	string				GetName()					{ return mName; }
	void				SetName(const string &n)	{ mName = n; }

	string				GetFilename()					{ return mFilename; }
	void				SetFilename(const string &n)	{ mFilename = n; }
	string				GetPath()						{ return mPath; }
	void				SetPath(const string &p)		{ mPath = p; }

	Model				*GetModel()						{ return mModel; }
	void				SetModel(Model *model);

	void				LoadLabels(const string &baseName = "");

	// -- Static --
	static DataSet*		ReadDataSet(const string &dsName, const string &path = "", const string &baseName = "");
	
	// Returns full path (path + filename)
	static string		DeterminePathFileExt(string &path, string &filename, string &extension);

protected:
	// Helpers
	void				CacheItemsets();

	// Attached model (if any)
	Model				*mModel;

	// Horizontal representation
	Row**				mRows;
	uint32				mNumRows;
	string*				mRowLabels;

	// Vertical representation
	Column**			mColumns;
	ColumnType*			mColumnTypes;
	uint32				mNumColumns;

	// Weights (for weighted covering)
	float				*mWeights;

	// Metadata
	string				mName;
	string				mFilename;
	string				mPath;

	// Cached itemsets
	ItemSet				**mItemSetCache;
};

#endif // __DATASET_H
