//
// Copyright (c) 2011-2012, Matthijs van Leeuwen
//
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
// Neither the name of the Universiteit Utrecht, KU Leuven, nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
#include "../../elobal.h"

#if defined (__unix__)
#include <glibc_s.h>
#endif

#include <Config.h>
#include <FileUtils.h>
#include <StringUtils.h>

#include <Bass.h>

#include "../data/BinaryColumn.h"
#include "../data/Row.h"
#include "../datasubset/DataSubset.h"

#include "EMMDataSet.h"

EMMDataSet::EMMDataSet() {
	mBinaryDescriptionSpace	= false;
	mBinaryModelSpace	= false;
	mNumDescriptionColumns	= 0;
	mNumModelColumns	= 0;
	mDescriptionColumns		= NULL;
	mModelColumns		= NULL;
}

EMMDataSet::EMMDataSet(EMMDataSet *ds) : DataSet(ds) {
	mBinaryModelSpace	= ds->mBinaryModelSpace;
	mBinaryDescriptionSpace	= ds->mBinaryDescriptionSpace;
	mNumDescriptionColumns	= ds->mNumDescriptionColumns;
	mNumModelColumns	= ds->mNumModelColumns;

	// Re-construct search and target column pointer arrays
	uint32 id;
	mDescriptionColumns = new Column *[mNumDescriptionColumns];
	for(uint32 i=0; i<mNumDescriptionColumns; i++) {
		id = ds->mDescriptionColumns[i]->GetId();
		uint32 j;
		for(j=0; j<mNumColumns; j++)
			if(mColumns[j]->GetId() == id)
				break;
		if(j == mNumColumns) THROW("Target column not found.");
		mDescriptionColumns[i] = mColumns[j];
	}
	mModelColumns = new Column *[mNumModelColumns];
	for(uint32 i=0; i<mNumModelColumns; i++) {
		id = ds->mModelColumns[i]->GetId();
		uint32 j;
		for(j=0; j<mNumColumns; j++)
			if(mColumns[j]->GetId() == id)
				break;
		if(j == mNumColumns) THROW("Target column not found.");
		mModelColumns[i] = mColumns[j];
	}
}
EMMDataSet::EMMDataSet(DataSet *dataset, string descriptionAtts, string modelAtts) : DataSet(dataset) {
	StringUtils::Trim(descriptionAtts, " "); StringUtils::Trim(modelAtts, " ");
	if(descriptionAtts.compare("*") == 0 && modelAtts.compare("*") == 0)
		THROW("Both description and model attribute space defined as * -- impossible.");
	
	if(descriptionAtts.compare("*") == 0 || modelAtts.compare("*") == 0) {		// one specified, find 'rest' for *
		bool descriptionAttsGiven = modelAtts.compare("*") == 0;

		uint32 numGiven;
		string *given = StringUtils::Tokenize(descriptionAttsGiven ? descriptionAtts : modelAtts, numGiven, ",", " ");
		Column **givenColumns = new Column *[numGiven];
		uint32 numRest = mNumColumns - numGiven;
		Column **restColumns = new Column *[numRest];

		uint32 gIdx = 0, rIdx = 0;
		for(uint32 i=0; i<mNumColumns; i++) {
			if(StringUtils::IsInArray(mColumns[i]->GetLabel(), given, numGiven)) {
				if(gIdx == numGiven) 
					THROW("More specified columns found than should be the case.");
				givenColumns[gIdx++] = mColumns[i];
			} else {
				if(rIdx == numRest) 
					THROW("More rest columns found than should be the case.");
				restColumns[rIdx++] = mColumns[i];
			}
		}
		if(gIdx != numGiven)
			THROW("Could not find all specified columns.");
		if(rIdx != numRest)
			THROW("Could not find all rest (*) columns.");
		
		if(descriptionAttsGiven) {
			mNumDescriptionColumns = numGiven;
			mDescriptionColumns = givenColumns;
			mNumModelColumns = numRest;
			mModelColumns = restColumns;
		} else {
			mNumDescriptionColumns = numRest;
			mDescriptionColumns = restColumns;
			mNumModelColumns = numGiven;
			mModelColumns = givenColumns;
		}

		delete[] given;

	} else {																// both specified

		// description attributes
		string *labels = StringUtils::Tokenize(descriptionAtts, mNumDescriptionColumns, ",", " ");
		mDescriptionColumns = new Column *[mNumDescriptionColumns];
		uint32 sIdx = 0;
		for(uint32 i=0; i<mNumColumns; i++) {
			if(StringUtils::IsInArray(mColumns[i]->GetLabel(), labels, mNumDescriptionColumns)) {
				if(sIdx == mNumDescriptionColumns) 
					THROW("More description columns found than should be the case.");
				mDescriptionColumns[sIdx++] = mColumns[i];
			}
		}
		delete[] labels;
		if(sIdx != mNumDescriptionColumns)
			THROW("Could not find all description columns.");

		// Target attributes
		labels = StringUtils::Tokenize(modelAtts, mNumModelColumns, ",", " ");
		mModelColumns = new Column *[mNumModelColumns];
		sIdx = 0;
		for(uint32 i=0; i<mNumColumns; i++) {
			if(StringUtils::IsInArray(mColumns[i]->GetLabel(), labels, mNumModelColumns)) {
				if(sIdx == mNumModelColumns) 
					THROW("More model columns found than should be the case.");
				mModelColumns[sIdx++] = mColumns[i];
			}
		}
		delete[] labels;
		if(sIdx != mNumModelColumns)
			THROW("Could not find all model columns.");
	}

	// Check whether spaces are binary-only
	mBinaryDescriptionSpace = true;
	for(uint32 i=0; i<mNumDescriptionColumns; i++)
		if(mDescriptionColumns[i]->GetType() != BinaryColumnType) {
			mBinaryDescriptionSpace = false;
			break;
		}
	mBinaryModelSpace = true;
	for(uint32 i=0; i<mNumModelColumns; i++)
		if(mModelColumns[i]->GetType() != BinaryColumnType) {
			mBinaryModelSpace = false;
			break;
		}

	ECHO(2, printf(" * EMM Dataset:\t\t%d description attributes, %d model attributes\n", mNumDescriptionColumns, mNumModelColumns));
	ECHO(2, printf(" * Binary description space:\t%s\n", mBinaryDescriptionSpace ? "yes" : "no"));
	ECHO(2, printf(" * Binary model space:\t%s\n", mBinaryModelSpace ? "yes" : "no"));
}
EMMDataSet::~EMMDataSet() {
	if(mDescriptionColumns != NULL)
		delete[] mDescriptionColumns;
	if(mModelColumns != NULL)
		delete[] mModelColumns;
}

DataSet* EMMDataSet::Clone() {
	return new EMMDataSet(this);
}
Database* EMMDataSet::CreateDescriptionDatabase() {
	if(!mBinaryDescriptionSpace)
		THROW("Dataset does not have a binary description space, cannot create (binary, fic) database.");
	return DataSet::CreateDatabase(mDescriptionColumns, mNumDescriptionColumns);
}
Database* EMMDataSet::CreateModelDatabase() {
	if(!mBinaryModelSpace)
		THROW("Dataset does not have a binary model space, cannot create (binary, fic) database.");
	return DataSet::CreateDatabase(mModelColumns, mNumModelColumns);
}
DataSet* EMMDataSet::CreateDescriptionDataSet() {
	return new DataSet(mName, mDescriptionColumns, mNumDescriptionColumns);
}
DataSet* EMMDataSet::CreateModelDataSet() {
	return new DataSet(mName, mModelColumns, mNumModelColumns);
}
EMMDataSet*	EMMDataSet::RetrieveDataSet(const string &dsName) {
	string baseName = dsName;
	if(dsName.find_first_of("[") != string::npos)
		baseName = dsName.substr(0, dsName.find_first_of("["));
	string path = Bass::GetDbReposDir() + baseName + "/";

	return EMMDataSet::ReadDataSet(dsName, path, baseName);
}
EMMDataSet*	EMMDataSet::ReadDataSet(const string &dsName, const string &path /* = "" */, const string &baseName /*= ""*/) {
	// Read DataSet
	DataSet *ds = DataSet::ReadDataSet(dsName, path, baseName);

	// Read .emm config
	Config *emmConfig = new Config();
	string emmFile = ds->GetPath() + ds->GetName() + ".emm";
	if(FileUtils::FileExists(emmFile)) {
		emmConfig->ReadFile(emmFile);
	} else {
		delete ds;
		delete emmConfig;
		THROW(".emm config file does not exist: " + emmFile);
	}

	// Build EMMDataSet
	if(!emmConfig->KeyExists("descriptionAtts") || !emmConfig->KeyExists("modelAtts")) {
		delete ds;
		delete emmConfig;
		THROW("Not a valid .emm config file.");
	}
	EMMDataSet *eds = new EMMDataSet(ds, emmConfig->Read<string>("descriptionAtts"), emmConfig->Read<string>("modelAtts"));

	// Cleanup
	delete ds;
	delete emmConfig;

	return eds;
}
