//
// Copyright (c) 2011-2012, Matthijs van Leeuwen
//
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
// Neither the name of the Universiteit Utrecht, KU Leuven, nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
#include "../../elobal.h"

#if defined (__unix__)
#include <glibc_s.h>
#endif

#include <Bass.h>

#include "../data/BinaryColumn.h"
#include "../data/Row.h"
#include "../dataset/DataSet.h"

#include "BinarySubset.h"

BinarySubset::BinarySubset(DataSet *ds) : DataSubset(ds) {
	mSubsetMask = new bool[ds->GetNumRows()];
	for(uint32 i=0; i<ds->GetNumRows(); i++)
		mSubsetMask[i] = false;
}
BinarySubset::BinarySubset(DataSet *ds, bool *mask, bool takeOwnership) : DataSubset(ds) {
	uint32 numRows = ds->GetNumRows();
	if(takeOwnership) {
		mSubsetMask = mask;
	} else {
		mSubsetMask = new bool[numRows];
		memcpy(mSubsetMask, mask, numRows * sizeof(bool));
	}
	for(uint32 i=0; i<numRows; i++)
		if(mSubsetMask[i])
			mSize++;
}
BinarySubset::BinarySubset(BinarySubset *datasubset) : DataSubset(datasubset) {
	mSubsetMask = new bool[mDataSet->GetNumRows()];
	memcpy(mSubsetMask, datasubset->mSubsetMask, mDataSet->GetNumRows() * sizeof(bool));
}
BinarySubset::~BinarySubset() {
	delete[] mSubsetMask;
}
DataSubset *BinarySubset::Clone() {
	return new BinarySubset(this);
}
void BinarySubset::AddToSubset(const uint32 row) {
	if(!mSubsetMask[row]) { // only add if not already there
		mSubsetMask[row] = true;
		++mSize;
	}
}
bool BinarySubset::IsInSubset(const uint32 row) {
	return mSubsetMask[row];
}
void BinarySubset::RemoveFromSubset(const uint32 row) {
	if(mSubsetMask[row]) { // only remove if there
		mSubsetMask[row] = false;
		--mSize;
	}
}
bool BinarySubset::Equals(DataSubset *subset) {
	bool *otherMask = subset->GetMask();
	for(uint32 r=0; r<mDataSet->GetNumRows(); r++)
		if(mSubsetMask[r] != otherMask[r])
			return false;
	return true;
}
uint32* BinarySubset::GetRowsInSubset() {
	uint32 *rows = new uint32[mSize];
	uint32 rIdx = 0;
	for(uint32 r=0; r<mDataSet->GetNumRows(); r++)
		if(mSubsetMask[r])
			rows[rIdx++] = r;
	if(rIdx != mSize)
		THROW("Number of 1s in subset mask does not match size of binary subset.");
	return rows;
}
uint32* BinarySubset::GetRowIdsInSubset() {
	Row **rows = mDataSet->GetRows();
	uint32 *rowIds = new uint32[mSize];
	uint32 rIdx = 0;
	for(uint32 r=0; r<mDataSet->GetNumRows(); r++)
		if(mSubsetMask[r])
			rowIds[rIdx++] = rows[r]->GetId();
	if(rIdx != mSize)
		THROW("Number of 1s in subset mask does not match size of binary subset.");
	return rowIds;
}

BinaryColumn* BinarySubset::GetBinaryColumn() {
	return new BinaryColumn(UINT32_MAX_VALUE, mDataSet->GetNumRows(), mSubsetMask, "subset");
}
void BinarySubset::UseThisMask(bool *mask, const uint32 numOnes) {
	delete[] mSubsetMask;
	mSubsetMask = mask;
	mSize = numOnes;
}
