//
// Copyright (c) 2011-2012, Matthijs van Leeuwen
//
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
// Neither the name of the Universiteit Utrecht, KU Leuven, nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
#include "../../elobal.h"

#if defined (__unix__)
#include <glibc_s.h>
#endif

#include <Bass.h>

#include "../data/BinaryColumn.h"
#include "../data/Row.h"
#include "../dataset/DataSet.h"

#include "Uint32ArraySubset.h"

Uint32ArraySubset::Uint32ArraySubset(DataSet *ds) : DataSubset(ds) {
	mSubsetArray = NULL;
}
Uint32ArraySubset::Uint32ArraySubset(DataSet *ds, bool *mask, bool takeOwnership) : DataSubset(ds) {
	uint32 numRows = ds->GetNumRows();
	mSize = 0;
	for(uint32 i=0; i<numRows; i++)
		if(mask[i])
			mSize++;
	mSubsetArray = new uint32[mSize];
	mSize = 0;
	for(uint32 i=0; i<numRows; i++)
		if(mask[i])
			mSubsetArray[mSize++] = i;

	if(takeOwnership)
		delete[] mask;
}
Uint32ArraySubset::Uint32ArraySubset(Uint32ArraySubset *datasubset) : DataSubset(datasubset) {
	mSubsetArray = new uint32[datasubset->GetSize()];
	memcpy(mSubsetArray, datasubset->mSubsetArray, datasubset->GetSize() * sizeof(uint32));
}
Uint32ArraySubset::~Uint32ArraySubset() {
	delete[] mSubsetArray;
}
DataSubset *Uint32ArraySubset::Clone() {
	return new Uint32ArraySubset(this);
}
void Uint32ArraySubset::AddToSubset(const uint32 row) {
	if(SearchInArray(row) == UINT32_MAX_VALUE) { // only add if not already there
		uint32 *arr = new uint32[mSize+1];
		uint32 idx=0;
		for(; idx<mSize && mSubsetArray[idx]<row; idx++)
			arr[idx] = mSubsetArray[idx];
		arr[idx++] = row;
		for(; idx<=mSize; idx++)
			arr[idx] = mSubsetArray[idx-1];
		delete[] mSubsetArray;
		++mSize;
	}
}
bool Uint32ArraySubset::IsInSubset(const uint32 row) {
	return SearchInArray(row) != UINT32_MAX_VALUE;
}
void Uint32ArraySubset::RemoveFromSubset(const uint32 row) {
	if(SearchInArray(row) != UINT32_MAX_VALUE) { // only remove if there
		uint32 *arr = new uint32[mSize-1];
		uint32 idx=0;
		for(; idx<mSize && mSubsetArray[idx]<row; idx++)
			arr[idx] = mSubsetArray[idx];
		idx++;
		for(; idx<mSize; idx++)
			arr[idx-1] = mSubsetArray[idx];
		delete[] mSubsetArray;
		--mSize;
	}
}
bool Uint32ArraySubset::Equals(DataSubset *subset) {
	if(subset->GetSize() != mSize)
		return false;
	uint32 *otherArray = subset->GetArray();
	for(uint32 r=0; r<mSize; r++)
		if(mSubsetArray[r] != otherArray[r])
			return false;
	return true;
}
uint32* Uint32ArraySubset::GetRowsInSubset() {
	uint32 *rows = new uint32[mSize];
	memcpy(rows, mSubsetArray, mSize*sizeof(uint32));
	return rows;
}
uint32* Uint32ArraySubset::GetRowIdsInSubset() {
	Row **rows = mDataSet->GetRows();
	uint32 *rowIds = new uint32[mSize];
	uint32 rIdx = 0;
	for(uint32 r=0; r<mSize; r++)
		rowIds[rIdx++] = rows[mSubsetArray[r]]->GetId();
	return rowIds;
}

BinaryColumn* Uint32ArraySubset::GetBinaryColumn() {
	uint32 numRows = mDataSet->GetNumRows();
	BinaryColumn *bcol = new BinaryColumn(UINT32_MAX_VALUE, numRows, "subset");

	uint32 idx = 0;
	for(uint32 r=0; r<numRows; r++)
		if(r == mSubsetArray[idx]) {
			bcol->SetValue(r, true);
			++idx;
		} else {
			bcol->SetValue(r, false);
		}
	return bcol;
}
uint32 Uint32ArraySubset::SearchInArray(const uint32 row) {
	uint32 minIdx = 0;
	uint32 maxIdx = mSize-1;
	uint32 tstIdx = UINT32_MAX_VALUE;
	bool found = false;
	while(minIdx+1 < maxIdx) { // while there are at least 3 possibilities
		tstIdx = minIdx + (maxIdx - minIdx)/2;
		if(mSubsetArray[tstIdx] > row) {
			maxIdx = tstIdx - 1;
		} else if(mSubsetArray[tstIdx] < row) {
			minIdx = tstIdx + 1;
		} else {
			return tstIdx;
		}
	}
	if(mSubsetArray[minIdx] == row)
		return minIdx;
	if(mSubsetArray[maxIdx] == row)
		return maxIdx;
	return UINT32_MAX_VALUE; // not found
}
