//
// Copyright (c) 2011-2012, Matthijs van Leeuwen
//
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
// Neither the name of the Universiteit Utrecht, KU Leuven, nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
#include "../../elobal.h"

#include <Bass.h>
#include <FileUtils.h>

#include <db/Database.h>
#include <itemstructs/ItemSet.h>

#include "Subgroup.h"
#include "../conditions/ConditionArray.h"
#include "../../models/Model.h"
#include "../../measures/QualityMeasure.h"

#include "SubgroupCollection.h"

SubgroupCollection::SubgroupCollection(uint32 maxSize /* = 0 */, const bool cacheSubsets /*= true*/) {
	mSubgroups = new SubgroupList();

	mMaxSize = maxSize;
	mCacheSubsets = cacheSubsets;
	mSize = 0;
	mNumCandidates = 0;
	mAutoSort = true;		// Only influences behaviour when maxSize = 0

	mMaxQuality = DOUBLE_MIN_VALUE;
	mMinQuality = DOUBLE_MAX_VALUE;
}
SubgroupCollection::~SubgroupCollection() {
	for(SubgroupList::iterator it = mSubgroups->begin(); it!=mSubgroups->end(); it++)
		delete (Subgroup *)(*it);
	delete mSubgroups;
}
bool SubgroupCollection::GoodEnough(double quality) {
	if(mMaxSize == 0) {				// Always add, no maxSize
		return true;

	} else {						// Maximum size
		if(mSize < mMaxSize)
			return true;	// maxSize not yet reached, always add
		else
			return quality > mMinQuality; // depends on quality
	}

}
void SubgroupCollection::Add(Subgroup *subgroup) {
	if(!mCacheSubsets)
		subgroup->ForgetSubset();

	if(mMaxSize == 0) {				// Always add, no maxSize
		if(mAutoSort)
			Insert(subgroup);
		else
			Pushback(subgroup);		// no sorting (only possible without maxSize)

	} else {
		if(mSize < mMaxSize)
			Insert(subgroup);
		else
			InsertAndTruncate(subgroup);
	}
}
void SubgroupCollection::Pushback(Subgroup *subgroup) {
	mSubgroups->push_back(subgroup->Clone());
	++mSize;
	double qual = subgroup->GetQuality();
	if(qual > mMaxQuality)
		mMaxQuality = qual;
	if(qual < mMinQuality)
		mMinQuality = qual;
}
void SubgroupCollection::Insert(Subgroup *subgroup) {
	double qual = subgroup->GetQuality();
	if(qual > mMaxQuality) {				// new best
		mMaxQuality = qual;
		if(mSize == 0)
			mMinQuality = qual;	// first one
		mSubgroups->push_front(subgroup->Clone());

	} else if(qual <= mMinQuality) {		// new worst
		mMinQuality = qual;
		mSubgroups->push_back(subgroup->Clone());

	} else {								// in the middle
		SubgroupList::iterator it = mSubgroups->begin();
		while((*it)->GetQuality() >= qual)
			++it;
		mSubgroups->insert(it, subgroup->Clone());
	}

	++mSize;
}
void SubgroupCollection::InsertAndTruncate(Subgroup *subgroup) {
	double qual = subgroup->GetQuality();

	if(qual > mMaxQuality) {				// new best
		mMaxQuality = qual;
		mSubgroups->push_front(subgroup->Clone());

	} else if(qual <= mMinQuality) {		// does not belong to topK
		return;

	} else {								// in the middle
		SubgroupList::iterator it = mSubgroups->begin();
		while((*it)->GetQuality() >= qual)
			++it;
		mSubgroups->insert(it, subgroup->Clone());
	}

	delete mSubgroups->back();
	mSubgroups->pop_back();
	mMinQuality = mSubgroups->back()->GetQuality();
}
void SubgroupCollection::CacheModels(QualityMeasure *measure) {
	for(SubgroupList::iterator it=mSubgroups->begin(); it!=mSubgroups->end(); ++it)
		measure->CacheModel(*it);
}
void SubgroupCollection::CacheSubsets(DataSet *dataSet) {
	for(SubgroupList::iterator it=mSubgroups->begin(); it!=mSubgroups->end(); ++it)
		(*it)->CacheSubset(dataSet);
}
void SubgroupCollection::GetUnrefinedSubgroupsIn(SubgroupList *subgroups) {
	for(SubgroupList::iterator it=mSubgroups->begin(); it!=mSubgroups->end(); ++it)
		if(!(*it)->IsRefined())
			subgroups->push_back(*it);
}
void SubgroupCollection::UpdateStats() {
	mSize = (uint32) mSubgroups->size();
	mMaxQuality = DOUBLE_MIN_VALUE;
	mMinQuality = DOUBLE_MAX_VALUE;
	for(SubgroupList::iterator it=mSubgroups->begin(); it!=mSubgroups->end(); ++it) {
		double qual = (*it)->GetQuality();
		if(qual > mMaxQuality)
			mMaxQuality = qual;
		if(qual < mMinQuality)
			mMinQuality = qual;
	}
}
void SubgroupCollection::SortOnQuality() {
	SubgroupList *sorted = new SubgroupList(); // perform simple insertion sort
	for(SubgroupList::iterator it=mSubgroups->begin(); it!=mSubgroups->end(); ++it) {
		double qual = (*it)->GetQuality();
		SubgroupList::iterator it2 = sorted->begin();
		while(it2!=sorted->end() && (*it2)->GetQuality() >= qual)
			++it2;
		sorted->insert(it2, *it);
	}
	delete mSubgroups;
	mSubgroups = sorted;
}
double SubgroupCollection::ComputeDiversityScore(const uint32 numRows) {
	double score = 0.0;

	// Look at coverage of each subgroup
	uint32 coverSum = 0;
	uint32 *counts = new uint32[numRows];
	for(uint32 i=0; i<numRows; i++)
		counts[i] = 0;
	for(SubgroupList::iterator it=mSubgroups->begin(); it!=mSubgroups->end(); ++it) {
		coverSum += (*it)->GetSize();
		bool *mask = (*it)->GetMask();
		for(uint32 i=0; i<numRows; i++)
			if(mask[i])
				++counts[i];
	}

	// Compute expected coverage for each row
	double expected = coverSum / (double)numRows;

	// Sum deviations for all rows and normalize
	for(uint32 i=0; i<numRows; i++)
		score += abs(expected - counts[i]) / expected;
	score /= numRows;

	// Cleanup and return result
	delete[] counts;
	return score;
}
SubgroupCollectionStats SubgroupCollection::ComputeStats(const uint32 numRows) {
	SubgroupCollectionStats stats;

	// Init
	stats.numSubgroups = (uint32) mSubgroups->size();
	stats.qualityMin = DOUBLE_MAX_VALUE;
	stats.qualityMax = DOUBLE_MIN_VALUE;
	stats.qualityAvg = 0.0;
	stats.qualitySd = 0.0;
	stats.sizeAvg = 0.0;
	stats.sizeSd = 0.0;
	stats.descAvg = 0.0;
	stats.descSd = 0.0;
	stats.diversityPenalty = 0.0;

	// Init temp
	uint32 numSubgroups = (uint32) mSubgroups->size();
	uint32 coverSum = 0;
	uint32 *coverCounts = new uint32[numRows];
	for(uint32 i=0; i<numRows; i++)
		coverCounts[i] = 0;
	ItemSet **coverSets = NULL;
	uint32 idx = 0;
	bool computeJE = numSubgroups < UINT16_MAX_VALUE;
	if(computeJE) {
		coverSets = new ItemSet *[numRows];
		for(uint32 i=0; i<numRows; i++)
			coverSets[i] = ItemSet::CreateEmpty(Uint16ItemSetType, numSubgroups);
	}

	// Look at each subgroup
	for(SubgroupList::iterator it=mSubgroups->begin(); it!=mSubgroups->end(); ++it) {
		double qual = (*it)->GetQuality();
		uint32 size = (*it)->GetSize();

		// simple stats
		if(qual < stats.qualityMin)
			stats.qualityMin = qual;
		if(qual > stats.qualityMax)
			stats.qualityMax = qual;
		stats.qualityAvg += qual;
		stats.sizeAvg += (double)size;
		stats.descAvg += (double)(*it)->GetConditions()->GetSize();

		if(computeJE) {
			// diversity score & joint entropy
			coverSum += size;
			bool *mask = (*it)->GetMask();
			for(uint32 i=0; i<numRows; i++)
				if(mask[i]) {
					++coverCounts[i];
					coverSets[i]->AddItemToSet(idx);
				}
			idx++;

		} else {
			// diversity score only
			coverSum += size;
			bool *mask = (*it)->GetMask();
			for(uint32 i=0; i<numRows; i++)
				if(mask[i])
					++coverCounts[i];
		}
	}

	// Compute averages
	stats.qualityAvg /= numSubgroups;
	stats.sizeAvg /= numSubgroups;
	stats.descAvg /= numSubgroups;

	// Compute stddevs
	for(SubgroupList::iterator it=mSubgroups->begin(); it!=mSubgroups->end(); ++it) {
		stats.qualitySd += pow((*it)->GetQuality() - stats.qualityAvg, 2.0);
		stats.sizeSd += pow((*it)->GetSize() - stats.sizeAvg, 2.0);
		stats.descSd += pow((*it)->GetConditions()->GetSize() - stats.descAvg, 2.0);
	}
	stats.qualitySd /= numSubgroups;
	stats.sizeSd /= numSubgroups;
	stats.descSd /= numSubgroups;
	stats.qualitySd = sqrt(stats.qualitySd);
	stats.sizeSd = sqrt(stats.sizeSd);
	stats.descSd = sqrt(stats.descSd);

	/// For diversity score (Cover Redundancy in paper)
	// Compute expected coverage for each row
	double expected = coverSum / (double)numRows;
	// Sum deviations for all rows and normalize
	for(uint32 i=0; i<numRows; i++)
		stats.diversityPenalty += abs(expected - coverCounts[i]) / expected;
	stats.diversityPenalty /= numRows;

	// Joint entropy
	if(computeJE) {
		Database *coverDB = new Database(coverSets, numRows, false);
		coverDB->SetAlphabetSize(numSubgroups);
		coverDB->ComputeEssentials();
		coverDB->Bin();

		double invNumRows = 1.0 / numRows;
		stats.jointEntropy = 0;
		for(uint32 r=0; r<coverDB->GetNumRows(); r++) {
			double p = coverDB->GetRow(r)->GetUsageCount() * invNumRows;
			stats.jointEntropy -= p * Log2(p);
		}
		delete coverDB;
	} else
		stats.jointEntropy = -1.0;

	// Cleanup and return result
	delete[] coverCounts;
	return stats;
}
void SubgroupCollection::WriteStatsFile(const string &filename, DataSet *descriptionDS) {
	// Open file
	FILE *fp;
	fopen_s(&fp, filename.c_str(), "w");
	
	// Write header
	fprintf_s(fp, "sg;quality;size;#conditions;description\n");

	// Write each subgroup
	uint32 idx=1;
	for(SubgroupList::iterator it=mSubgroups->begin(); it!=mSubgroups->end(); ++it)
		fprintf_s(fp, "%u;%.05f;%u;%u;%s\n", idx++, (*it)->GetQuality(), (*it)->GetSize(), (*it)->GetConditions()->GetSize(), (*it)->GetConditions()->ToString(descriptionDS).c_str());

	// Close file
	fclose(fp);
}
void SubgroupCollection::WriteSubgroups(const string &filename, DataSet *descriptionDS) {
	// Open file
	FILE *fp;
	fopen_s(&fp, filename.c_str(), "w");

	// Write each subgroup
	uint32 idx=1;
	for(SubgroupList::iterator it=mSubgroups->begin(); it!=mSubgroups->end(); ++it)
		fprintf_s(fp, "%s\n", (*it)->GetConditions()->ToString(descriptionDS).c_str());

	// Close file
	fclose(fp);
}
void SubgroupCollection::WriteSubsets(const string &prefix, bool outputSubsets /*= true*/, bool outputLabelledSubsets /*= false*/, string *labels /*= NULL*/) {
	string subsetDir = Bass::GetWorkingDir() + "subsets/";
	if(outputSubsets)
		FileUtils::CreateDirIfNotExists(subsetDir);
	string labelledSubsetDir = Bass::GetWorkingDir() + "labelledsubsets/";
	if(outputLabelledSubsets)
		FileUtils::CreateDirIfNotExists(labelledSubsetDir);
	string fullPath; char temp[10];

	uint32 idx=1;
	for(SubgroupList::iterator it=mSubgroups->begin(); it!=mSubgroups->end(); ++it) {
		sprintf_s(temp, 10, "%04u", idx++);
		DataSubset *subset = (*it)->GetSubset();
		if(outputSubsets) {
			fullPath = subsetDir + prefix + temp + ".txt";
			subset->Save(fullPath);
		}
		if(outputLabelledSubsets) {
			fullPath = labelledSubsetDir + prefix + temp + ".csv";
			subset->SaveLabelled(fullPath, labels);
		}
	}
}
void SubgroupCollection::WriteModels(const string &prefix) {
	string modelDir = Bass::GetWorkingDir() + "models/";
	FileUtils::CreateDirIfNotExists(modelDir);
	string fullPath; char temp[10];

	uint32 idx=1;
	for(SubgroupList::iterator it=mSubgroups->begin(); it!=mSubgroups->end(); ++it) {
		sprintf_s(temp, 10, "%04u", idx++);
		fullPath = modelDir + prefix + temp;
		Model *model = (*it)->GetModel();
		if(model!=NULL)
			model->WriteHumanReadable(fullPath); // Let model decide on extension
	}
}
