//
// Copyright (c) 2011-2012, Matthijs van Leeuwen
//
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
// Neither the name of the Universiteit Utrecht, KU Leuven, nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
#include "../elobal.h"

#include <Bass.h>
#include <Config.h>
#include <logger/Log.h>

#include "../datastructs/dataset/DataSet.h"
#include "../datastructs/subgroup/Subgroup.h"
#include "../models/Model.h"

#include "ChiSquaredMeasure.h"
#include "MeanTestMeasure.h"
#include "WRAccMeasure.h"
#include "WRAccNumeric.h"
#include "compression/KLQualityMeasure.h"
#include "compression/KrimpGainMeasure.h"

#include "QualityMeasure.h"

QualityMeasure::QualityMeasure(Config *config, DataSet *ds) {
	mDS = ds;
	mDSNumRows = ds->GetNumRows();
	mNumCols = ds->GetNumColumns();
	if(config->KeyExists("sgModels")) {
		if(config->Read<string>("sgModels").compare("rebuild")==0)
			mModelPolicy = RebuildModels;
		else
			mModelPolicy = CacheModels;
	} else
		mModelPolicy = CacheModels;
}
QualityMeasure::~QualityMeasure() {

}
void QualityMeasure::LogParameters() {
	LOG_MSG("Measure:\t\t" + GetName());
}
double QualityMeasure::ComputeWeightedQuality(Subgroup *subgroup, const bool computeOE /*= false*/) {
	if(computeOE && mDS->HasWeights())
		THROW("Cannot use optimistic estimate when weighting quality.");
	if(mDS->HasWeights()) {
		double q = mDS->ComputeSubsetWeight(subgroup->GetSubset()) * ComputeQuality(subgroup);
		subgroup->SetQuality(q);
		return q;
	}
	return ComputeQuality(subgroup, computeOE);
}
QualityMeasure* QualityMeasure::Create(Config *config, DataSet *ds) {
	string desc = config->Read<string>("measure");
	if(desc.compare("KG") == 0 || desc.compare("WKG") == 0)
		return new KrimpGainMeasure(config, ds);
	if(desc.compare("KL") == 0 || desc.compare("WKL") == 0)
		return new KLQualityMeasure(config, ds);
	if(desc.compare("WRAcc") == 0) {
		if(ds->AllColumnsNumeric())
			return new WRAccNumeric(config, ds);
		else
			return new WRAccMeasure(config, ds);
	}
	if(desc.compare("ChiSquared") == 0)
		return new ChiSquaredMeasure(config, ds);
	if(desc.compare("meantest") == 0)
		return new MeanTestMeasure(config, ds);
	THROW("Unknown quality measure type.");
}
void QualityMeasure::SaveToSubgroup(Subgroup *subgroup, Model *model, double quality) {
	if(mModelPolicy == CacheModels)
		subgroup->SetModel(model);
	else 
		delete model;
	subgroup->SetQuality(quality);
}
