//
// Copyright (c) 2011-2012, Matthijs van Leeuwen
//
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
// Neither the name of the Universiteit Utrecht, KU Leuven, nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
#include "../elobal.h"

#include <Bass.h>
#include <Config.h>
#include <logger/Log.h>

#include "../datastructs/dataset/DataSet.h"
#include "../datastructs/datasubset/DataSubset.h"
#include "../datastructs/data/BinaryColumn.h"
#include "../datastructs/data/FloatColumn.h"
#include "../datastructs/data/NominalColumn.h"
#include "../datastructs/subgroup/Subgroup.h"

#include "../models/ContingencyTable.h"

#include "WRAccMeasure.h"

WRAccMeasure::WRAccMeasure(Config *config, DataSet *ds) : QualityMeasure(config, ds) {
	// Configuration
	string mode = config->Read<string>("WRAccMode", "1vsAll");
	if(mode.compare("single") == 0)
		mMode = SingleBinaryWRAcc;
	else if(mode.compare("1vsAllWeighted") == 0)
		mMode = OnevsAllWeightedWRAcc;
	else if(mode.compare("1vs1") == 0)
		mMode = OnevsOneWRAcc;
	else 
		mMode = OnevsAllWRAcc;

	// -- Build model for overall dataset --
	mDSModel = ContingencyTable::Build(ds);
	ds->SetModel(mDSModel);
}
WRAccMeasure::~WRAccMeasure() {
	// mDSModel is taken over by DS
}
void WRAccMeasure::LogParameters() {
	QualityMeasure::LogParameters();
	if(mMode == SingleBinaryWRAcc)
		LOG_MSG("WRAcc mode:\t\tsingle binary (traditional)");
	else if(mMode == OnevsAllWRAcc)
		LOG_MSG("WRAcc mode:\t\t1-vs-all");
	else if(mMode == OnevsAllWeightedWRAcc)
		LOG_MSG("WRAcc mode:\t\tweigthed 1-vs-all");
	else if(mMode == OnevsOneWRAcc)
		LOG_MSG("WRAcc mode:\t\t1-vs-1");
	else
		LOG_MSG("WRAcc mode:\t\timpossible.");
}

double WRAccMeasure::ComputeQuality(Subgroup *subgroup, const bool computeOE /*= false*/) {
	// Init
	DataSubset *subset = subgroup->GetSubset();

	// Build ContingencyTable
	ContingencyTable *model = ContingencyTable::Build(mDS, subset);

	// Compute and sum WRAccs
	double quality = 0.0;
	switch(mMode) {
		case SingleBinaryWRAcc:
			quality = ComputeSingle(model, subset->GetSize());
			if(computeOE)
				subgroup->SetOptimisticEstimate( ComputeSingleOE(model, subset->GetSize()) );
			break;
		case OnevsAllWRAcc:
			quality = ComputeOneVsAll(model, subset->GetSize());
			break;
		case OnevsAllWeightedWRAcc:
			quality = ComputeOneVsAllWeighted(model, subset->GetSize());
			break;
		case OnevsOneWRAcc:
			quality = ComputeOneVsOne(model, subset->GetSize());
			break;
	}

	// Save to subgroup
	SaveToSubgroup(subgroup, model, quality);

	// Return quality
	return quality;
}
void WRAccMeasure::CacheModel(Subgroup *subgroup) {
	subgroup->SetModel(ContingencyTable::Build(mDS, subgroup->GetSubset()));
}
double WRAccMeasure::ComputeSingle(ContingencyTable *sgModel, const uint32 sgSize) {
	// If multiple target columns specified, only look at the first one. Assume second value is binary true.
	// n/N * (p - p0)
	uint32 *sgCounts = sgModel->GetCountsForColumn(0);
	uint32 *dbCounts = mDSModel->GetCountsForColumn(0);
	double divider = (double)mDSNumRows;
	return (sgCounts[1] / divider) - ((sgSize * dbCounts[1]) / (divider * divider));
}
double WRAccMeasure::ComputeSingleOE(ContingencyTable *sgModel, const uint32 sgSize) {
	// n/N * p * (1 - p0)
	uint32 *sgCounts = sgModel->GetCountsForColumn(0);
	float *dbShares = mDSModel->GetSharesForColumn(0);
	return (sgCounts[1] / (double)mDSNumRows) * (1.0 - dbShares[1]);
}
double WRAccMeasure::ComputeOneVsAll(ContingencyTable *sgModel, const uint32 sgSize) {
	double wracc = 0.0;
	for(uint32 c=0; c<mNumCols; c++) {
		uint32 numVals = sgModel->GetNumValuesForColumn(c);
		uint32 *sgCounts = sgModel->GetCountsForColumn(c);
		uint32 *dbCounts = mDSModel->GetCountsForColumn(c);
		int32 acc = 0;
		for(uint32 i=0; i<numVals; i++)
			acc += abs((int32)sgCounts[i]*(int32)mDSNumRows - (int32)sgSize*(int32)dbCounts[i]);
		wracc += acc / (double)(numVals*mDSNumRows*mDSNumRows);
	}
	wracc /= mNumCols;
	return wracc;
}
double WRAccMeasure::ComputeOneVsAllWeighted(ContingencyTable *sgModel, const uint32 sgSize) {
	double wracc = 0.0;
	for(uint32 c=0; c<mNumCols; c++) {
		uint32 numVals = sgModel->GetNumValuesForColumn(c);
		uint32 *sgCounts = sgModel->GetCountsForColumn(c);
		uint32 *dbCounts = mDSModel->GetCountsForColumn(c);
		int32 acc = 0;
		for(uint32 i=0; i<numVals; i++)
			acc += (int32)dbCounts[i] * abs((int32)sgCounts[i]*(int32)mDSNumRows - (int32)sgSize*(int32)dbCounts[i]);
		wracc += acc / (double)(mDSNumRows*mDSNumRows*mDSNumRows);
	}
	wracc /= mNumCols;
	return wracc;
}
double WRAccMeasure::ComputeOneVsOne(ContingencyTable *sgModel, const uint32 sgSize) {
	double wracc = 0.0, acc;
	for(uint32 c=0; c<mNumCols; c++) {
		uint32 numVals = sgModel->GetNumValuesForColumn(c);
		uint32 *sgCounts = sgModel->GetCountsForColumn(c);
		uint32 *dbCounts = mDSModel->GetCountsForColumn(c);
		acc = 0.0;
		for(uint32 i=0; i<numVals; i++)
			for(uint32 j=i+1; j<numVals; j++) {
				if(sgCounts[i]==0 && sgCounts[j]==0)
					continue;
				acc += abs((sgCounts[i]+sgCounts[j])/(double)(dbCounts[i]+dbCounts[j]) * ((double)sgCounts[i]/(sgCounts[i]+sgCounts[j]) - (double)dbCounts[i]/(dbCounts[i]+dbCounts[j])));
			}
		acc *= 2;
		acc /= numVals * (numVals - 1);
		wracc += acc;
	}
	wracc /= mNumCols;
	return wracc;
}
