//
// Copyright (c) 2011-2012, Matthijs van Leeuwen
//
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
// Neither the name of the Universiteit Utrecht, KU Leuven, nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
#include "../../elobal.h"

#include <Bass.h>
#include <Config.h>
#include <logger/Log.h>

#include <algo/CodeTable.h>
#include <db/Database.h>
#include <FicMain.h>

#include "../../datastructs/dataset/DataSet.h"
#include "../../datastructs/datasubset/DataSubset.h"
#include "../../datastructs/subgroup/Subgroup.h"
#include "../../models/compression/CodeTableWrap.h"

#include "KrimpGainMeasure.h"

KrimpGainMeasure::KrimpGainMeasure(Config *config, DataSet *ds) : CompressionBasedMeasure(config, ds) {
	// Configuration
	char firstChar = config->Read<string>("measure").at(0);
	mWeightedVersion =  firstChar == 'W' || firstChar == 'w';

	// Init config copy -- used for Krimping
	mKrimpConfig = new Config(config);

	// Construct binary database from dataset
	mDatabase = ds->CreateDatabase();

	// -- Build CT for overall dataset --
	mKrimpConfig->Set("iscName", mKrimpConfig->Read<string>("dbKrimp"));
	mFullCT = CreateCodeTable(mDatabase);
	mFullCT->AddOneToEachUsageCount();

	// Create model (wrap around fic://CT)
	mDSModel = new CodeTableWrap(mFullCT, mDatabase);
	ds->SetModel(mDSModel);

	// Prepare config for subgroup Krimping
	mKrimpConfig->Set("iscName", mKrimpConfig->Read<string>("sgKrimp"));
}
KrimpGainMeasure::~KrimpGainMeasure() {
	// mFullCT and mDatabase are taken over by mDSModel
	// mDSModel is taken over by DS
	delete mKrimpConfig;
}
double KrimpGainMeasure::ComputeQuality(Subgroup *subgroup, const bool computeOE /*= false*/) {
	// Init
	DataSubset *subset = subgroup->GetSubset();

	// Build subgroup DB & CT
	Database *sgDB = mDatabase->Subgroup(subset->GetMask(), subset->GetSize());
	CodeTable *sgCT = CreateCodeTable(sgDB);
	sgCT->AddOneToEachUsageCount();
	sgCT->CalcTotalSize(sgCT->GetCurStats());

	// Compute gain
	double quality = mFullCT->CalcEncodedDbSize(sgDB) - sgCT->GetCurStats().encDbSize;
	if(!mWeightedVersion) // !weighted = bits per transactions on average; weighted = total number of bits for the subgroup
		quality /= sgDB->GetNumTransactions();
	else if(mFullCT->GetCurStats().encDbSize != 0.0)
		quality /= mFullCT->GetCurStats().encDbSize;

	// Save to subgroup
	SaveToSubgroup(subgroup, new CodeTableWrap(sgCT, sgDB), quality); // subgroup takes model takes CT and DB

	// Return quality
	return quality;
}
void KrimpGainMeasure::CacheModel(Subgroup *subgroup) {
	DataSubset *subset = subgroup->GetSubset();
	Database *sgDB = mDatabase->Subgroup(subset->GetMask(), subset->GetSize());
	CodeTable *sgCT = CreateCodeTable(sgDB);
	sgCT->AddOneToEachUsageCount();
	sgCT->CalcTotalSize(sgCT->GetCurStats());
	subgroup->SetModel(new CodeTableWrap(sgCT, sgDB));
}
CodeTable *KrimpGainMeasure::CreateCodeTable(Database *db) {
	uint8 ol = Bass::GetOutputLevel(); 
	Bass::SetOutputLevel(0);
	CodeTable *ct = FicMain::CreateCodeTable(db, mKrimpConfig);
	Bass::SetOutputLevel(ol);
	return ct;
}
