//
// Copyright (c) 2011-2012, Matthijs van Leeuwen
//
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
// Neither the name of the Universiteit Utrecht, KU Leuven, nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
#include "../elobal.h"

#include <Bass.h>
#include <Config.h>
#include <logger/Log.h>

#include "../datastructs/data/BinaryColumn.h"
#include "../datastructs/data/NominalColumn.h"
#include "../datastructs/dataset/DataSet.h"
#include "../datastructs/datasubset/DataSubset.h"

#include "ContingencyTable.h"

ContingencyTable::ContingencyTable(const uint32 numCols, uint32 *numValues, uint32 **counts, uint32 numRows) {
	mNumColumns = numCols;
	mNumValues = numValues;
	mCounts = counts;

	if(numCols > 0) {
		mShares = new float *[mNumColumns];
		for(uint32 i=0; i<mNumColumns; i++) {
			mShares[i] = new float[mNumValues[i]];
			for(uint32 v=0; v<mNumValues[i]; v++)
				mShares[i][v] = mCounts[i][v] / (float)numRows;
		}
	}
}
ContingencyTable::ContingencyTable(ContingencyTable *ctable) {
	mNumColumns = ctable->mNumColumns;
	mNumValues = new uint32[mNumColumns];
	memcpy(mNumValues, ctable->mNumValues, mNumColumns * sizeof(uint32));

	mCounts = new uint32 *[mNumColumns];
	mShares = new float *[mNumColumns];
	for(uint32 i=0; i<mNumColumns; i++) {
		mCounts[i] = new uint32[mNumValues[i]];
		mShares[i] = new float[mNumValues[i]];
		memcpy(mCounts[i], ctable->mCounts[i], mNumValues[i] * sizeof(uint32));
		memcpy(mShares[i], ctable->mShares[i], mNumValues[i] * sizeof(float));
	}
}
ContingencyTable::ContingencyTable() {
	mNumColumns = 0;
	mNumValues = NULL;
	mCounts = NULL;
	mShares = NULL;
}
ContingencyTable::~ContingencyTable() {
	if(mNumColumns > 0) {
		for(uint32 i=0; i<mNumColumns; i++) {
			delete[] mCounts[i];
			delete[] mShares[i];
		}
		delete[] mCounts;
		delete[] mShares;
		delete[] mNumValues;
	}
}
Model *ContingencyTable::Clone() {
	return new ContingencyTable(this);
}
void ContingencyTable::WriteHumanReadable(const string &filename) {
	// Open file
	string fullPath = filename + ".csv";
	FILE *fp;
	fopen_s(&fp, fullPath.c_str(), "w");

	// Write header
	fprintf_s(fp, "Contingency table (absolute counts)\ncolumn;");
	uint32 maxNum = 0;
	for(uint32 i=0; i<mNumColumns; i++)
		if(mNumValues[i] > maxNum)
			maxNum = mNumValues[i];
	for(uint32 i=0; i<maxNum; i++)
		fprintf_s(fp, "value%u;", i);
	fprintf_s(fp, "\n");

	// For each column
	for(uint32 i=0; i<mNumColumns; i++) {
		fprintf_s(fp, "%u;", i);
		for(uint32 v=0; v<mNumValues[i]; v++)
			fprintf_s(fp, "%u;", mCounts[i][v]);
		fprintf_s(fp, "\n");
	}
		
	// Write header
	fprintf_s(fp, "\nContingency table (relative)\ncolumn;");
	for(uint32 i=0; i<maxNum; i++)
		fprintf_s(fp, "value%u;", i);
	fprintf_s(fp, "\n");

	// For each column
	for(uint32 i=0; i<mNumColumns; i++) {
		fprintf_s(fp, "%u;", i);
		for(uint32 v=0; v<mNumValues[i]; v++)
			fprintf_s(fp, "%.03f;", mShares[i][v]);
		fprintf_s(fp, "\n");
	}

	// Close file
	fclose(fp);
}
ContingencyTable *ContingencyTable::Build(DataSet *ds) {
	uint32 numColumns = ds->GetNumColumns();
	uint32 *numValues = new uint32[numColumns];
	uint32 **counts = new uint32 *[numColumns];
	uint32 numRows = ds->GetNumRows();

	// Foreach (independent) column
	Column **columns = ds->GetColumns();
	for(uint32 i=0; i<numColumns; i++) {
		switch(columns[i]->GetType()) {
			case BinaryColumnType:
				{
					numValues[i] = 2;
					counts[i] = new uint32[2];
					BinaryColumn *bCol = (BinaryColumn *)(columns[i]);
					bool *bValues = bCol->GetBoolValues();
					counts[i][1] = 0;
					for(uint32 r=0; r<numRows; r++)
						if(bValues[r])
							++counts[i][1];
					counts[i][0] = numRows - counts[i][1];
				}
				break;
			case FloatColumnType:
				THROW("Float columns unsupported by ContingencyTable.");
				break;
			case NominalColumnType:
				{
					NominalColumn *nCol = (NominalColumn *)(columns[i]);
					numValues[i] = nCol->GetNumValues();
					counts[i] = new uint32[numValues[i]];
					uint32 *nValues = nCol->GetNominalValues();
					for(uint32 n=0; n<numValues[i]; n++)
						counts[i][n] = 0;
					for(uint32 r=0; r<numRows; r++)
						++counts[i][nValues[r]];
				}
				break;
			default:
				THROW("Unsupported column type?!");
		}
	}
	return new ContingencyTable(numColumns, numValues, counts, numRows);
}
ContingencyTable *ContingencyTable::Build(DataSet *ds, DataSubset *subset) {
	uint32 numRows = subset->GetSize();
	uint32 *rows = subset->GetRowsInSubset();
	uint32 numColumns = ds->GetNumColumns();
	uint32 *numValues = new uint32[numColumns];
	uint32 **counts = new uint32 *[numColumns];

	// Foreach (independent) column
	Column **columns = ds->GetColumns();
	for(uint32 i=0; i<numColumns; i++) {
		switch(columns[i]->GetType()) {
		case BinaryColumnType:
			{
				numValues[i] = 2;
				counts[i] = new uint32[2];
				BinaryColumn *bCol = (BinaryColumn *)(columns[i]);
				bool *bValues = bCol->GetBoolValues();
				counts[i][1] = 0;
				for(uint32 r=0; r<numRows; r++)
					if(bValues[rows[r]])
						++counts[i][1];
				counts[i][0] = numRows - counts[i][1];
			}
			break;
		case FloatColumnType:
			THROW("Float columns unsupported by ContingencyTable.");
			break;
		case NominalColumnType:
			{
				NominalColumn *nCol = (NominalColumn *)(columns[i]);
				numValues[i] = nCol->GetNumValues();
				counts[i] = new uint32[numValues[i]];
				uint32 *nValues = nCol->GetNominalValues();
				for(uint32 n=0; n<numValues[i]; n++)
					counts[i][n] = 0;
				for(uint32 r=0; r<numRows; r++)
					++counts[i][nValues[rows[r]]];
			}
			break;
		default:
			THROW("Unsupported column type?!");
		}
	}

	delete[] rows;

	return new ContingencyTable(numColumns, numValues, counts, numRows);
}
