//
// Copyright (c) 2011-2012, Matthijs van Leeuwen
//
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
// Neither the name of the Universiteit Utrecht, KU Leuven, nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
#include "../../elobal.h"

#include <Bass.h>
#include <Config.h>
#include <logger/Log.h>

#include "../../datastructs/data/BinaryColumn.h"
#include "../../datastructs/data/NominalColumn.h"
#include "../../datastructs/dataset/DataSet.h"
#include "../../datastructs/datasubset/DataSubset.h"
#include "../ContingencyTable.h"

#include "KLCompressor.h"

KLCompressor::KLCompressor(DataSet *ds) {
	mTable = ContingencyTable::Build(ds);
}
KLCompressor::KLCompressor(DataSet *ds, DataSubset *subset) {
	mTable = ContingencyTable::Build(ds, subset);
}
KLCompressor::KLCompressor(KLCompressor *klc) {
	mTable = new ContingencyTable(klc->mTable);
}
KLCompressor::~KLCompressor() {
	delete mTable;
}
Model *KLCompressor::Clone() {
	return new KLCompressor(this);
}
void KLCompressor::GetEncodedTransactionSizes(DataSet *ds, float *sizes, DataSubset *subset /*= NULL*/) {
	// Set each size to 0 (iff indicated by subset!)
	if(subset==NULL) {
		for(uint32 r=0; r<ds->GetNumRows(); r++)
			sizes[r] = 0.0f;
	} else {
		for(uint32 r=0; r<ds->GetNumRows(); r++)
			if(subset->IsInSubset(r))
				sizes[r] = 0.0f;
	}

	// Get countsum for the data(sub)set this compressor was built upon
	uint32 countSum = subset==NULL ? ds->GetNumRows() : subset->GetSize();

	// Foreach (independent) column
	Column **columns = ds->GetColumns();
	for(uint32 c=0; c<ds->GetNumColumns(); c++) {
		// Cache code lengths
		uint32 *counts = mTable->GetCountsForColumn(c);
		uint32 numValues = mTable->GetNumValuesForColumn(c);
		float *lengths = new float[numValues];
		for(uint32 i=0; i<numValues; i++)
			lengths[i] = -1.0f * (float)Log2(counts[i] / (float)countSum); // without Laplace
//			lengths[i] = -1.0f * (float)Log2((counts[i]+1) / (float)(countSum+numValues)); Laplace

		// Go through the data
		switch(columns[c]->GetType()) {
			case BinaryColumnType:
				{
					BinaryColumn *bCol = (BinaryColumn *)(columns[c]);
					bool *bValues = bCol->GetBoolValues();
					if(subset == NULL) {
						for(uint32 r=0; r<ds->GetNumRows(); r++)
							sizes[r] += lengths[bValues[r] ? 1 : 0];
					} else {
						for(uint32 r=0; r<ds->GetNumRows(); r++)
							if(subset->IsInSubset(r))
								sizes[r] += lengths[bValues[r] ? 1 : 0];
					}
				}
				break;
			case FloatColumnType:
				THROW("Float columns unsupported by KLCompressor.");
				break;
			case NominalColumnType:
				{
					NominalColumn *nCol = (NominalColumn *)(columns[c]);
					uint32 *nValues = nCol->GetNominalValues();
					if(subset == NULL) {
						for(uint32 r=0; r<ds->GetNumRows(); r++)
							sizes[r] += lengths[nValues[r]];
					} else {
						for(uint32 r=0; r<ds->GetNumRows(); r++)
							if(subset->IsInSubset(r))
								sizes[r] += lengths[nValues[r]];
					}
				}
				break;
			default:
				THROW("Unsupported column type?!");
		}
		delete[] lengths;
	}
}
double KLCompressor::GetEncodedSize() {
	double size = 0.0;
	for(uint32 c=0; c<mTable->GetNumColumns(); c++) {
		uint32 *counts = mTable->GetCountsForColumn(c);
		float *probs = mTable->GetSharesForColumn(c);
		uint32 numProbs = mTable->GetNumValuesForColumn(c);
		for(uint32 i=0; i<numProbs; i++)
			if(counts[i] != 0)
				size -= counts[i] * Log2(probs[i]);
	}
	return size;
}
double KLCompressor::ComputeKLDivergence(KLCompressor *toModel) {
	double divergence = 0.0;
	for(uint32 c=0; c<mTable->GetNumColumns(); c++)
		divergence += ComputeKLDivergence(mTable->GetSharesForColumn(c), toModel->mTable->GetSharesForColumn(c), toModel->mTable->GetNumValuesForColumn(c));
	return divergence;
}
double KLCompressor::ComputeKLDivergence(float *fromProbs, float *toProbs, const uint32 numProbs) {
	double kl = 0.0f;
	for(uint32 i=0; i<numProbs; i++)
		if(toProbs[i] != 0.0f && fromProbs[i] != 0.0f)
			kl += toProbs[i] * Log2(toProbs[i] / fromProbs[i]);
	return kl;
}
void KLCompressor::WriteHumanReadable(const string &filename) {
	mTable->WriteHumanReadable(filename);
}
