//
// Copyright (c) 2011-2012, Matthijs van Leeuwen
//
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
// Neither the name of the Universiteit Utrecht, KU Leuven, nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
#include "../elobal.h"

#include <Bass.h>
#include <Config.h>

#include "../datastructs/conditions/ConditionArray.h"
#include "../datastructs/datasubset/DataSubset.h"
#include "../datastructs/subgroup/Subgroup.h"
#include "../datastructs/subgroup/SubgroupCollection.h"
#include "../measures/QualityMeasure.h"
#include "../models/Model.h"

#include "DominancePruner.h"

DominancePruner::DominancePruner() {

}
DominancePruner::~DominancePruner() {

}
void DominancePruner::Process(SubgroupCollection *collection, DataSet *ds, QualityMeasure *measure) {
	SubgroupList *subgroups = collection->GetSubgroups();

	// Prune each separate subgroup
	for(SubgroupList::iterator it=subgroups->begin(); it!=subgroups->end(); it++)
		Prune(*it, ds, measure);

	// Test subgroups for equal descriptions
	for(SubgroupList::iterator it=subgroups->begin(); it!=subgroups->end(); it++) {
		// Compare to all other subgroups
		for(SubgroupList::iterator it2=it; it2!=subgroups->end(); ) {
			if(it2==it) {
				++it2;
				continue; // skip itself
			}

			if((*it)->GetConditions()->Equals((*it2)->GetConditions())) {
				delete *it2;
				SubgroupList::iterator del = it2;
				++it2;
				subgroups->erase(del);
			} else {
				++it2;
			}
		}
	}

	collection->UpdateStats();
}
void DominancePruner::Prune(Subgroup *subgroup, DataSet *ds, QualityMeasure *measure) {
	// Init
	ConditionArray *bestConditions = subgroup->GetConditions();
	if(bestConditions->GetSize() <= 1)
		return;
	DataSubset *bestSubset = subgroup->GetSubset();
	double bestQuality = subgroup->GetQuality();
	Model *bestModel = subgroup->GetModel();
	if(bestModel != NULL)
		bestModel = bestModel->Clone();

	// Test each individual condition
	bool change;
	do {
		change = false;
		for(uint32 i=0; i<bestConditions->GetSize(); i++) {
			ConditionArray *testConditions = new ConditionArray(bestConditions, i); // remove 
			DataSubset *testSubset = testConditions->Apply(ds);
			subgroup->SetSubset(testSubset);
			double testQuality = measure->ComputeQuality(subgroup);
			if(testQuality >= bestQuality) { 
				delete bestConditions;		// replace best with test
				delete bestSubset;
				delete bestModel;
				bestConditions = testConditions;
				bestSubset = testSubset;
				bestQuality = testQuality;
				if(bestModel != NULL)
					bestModel = subgroup->GetModel()->Clone();
				change = true;
				break;

			} else {	// proceed with next
				delete testConditions;
				delete testSubset;
			}
		}
	} while(change && bestConditions->GetSize() > 1);

	// Give result to subgroup
	subgroup->SetConditions(bestConditions);
	subgroup->SetSubset(bestSubset);
	subgroup->SetQuality(bestQuality);
	if(bestModel != NULL)
		subgroup->SetModel(bestModel);
}
