//
// Copyright (c) 2011-2012, Matthijs van Leeuwen
//
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
// Neither the name of the Universiteit Utrecht, KU Leuven, nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
#include "../elobal.h"

#if defined (__unix__)
#include <glibc_s.h>
#endif

#include <Config.h>
#include <logger/Log.h>

#include "../datastructs/conditions/ConditionArray.h"
#include "../datastructs/dataset/DataSet.h"
#include "../datastructs/subgroup/Subgroup.h"
#include "../datastructs/subgroup/SubgroupCollection.h"
#include "../measures/QualityMeasure.h"

#include "IterativeSearch.h"

IterativeSearch::IterativeSearch(const float weight, const string &searchDesc, Config *config, DataSet *descDS, DataSet *modelDS) : Search(config, descDS, modelDS) {
	mWeight = weight;

	mUpdatedDescDataSet = descDS->Clone();
	mUpdatedModelDataSet = modelDS->Clone();

	mSearchDesc = searchDesc;
	mSearchConfig = new Config(config);
	mSearchConfig->Set("topK", 1); // find only one pattern at a time
	mSearch = Create(mSearchDesc, mSearchConfig, mUpdatedDescDataSet, mUpdatedModelDataSet);

	if(mWeight != 0.0) {			// Init weighted covering
		mUpdatedModelDataSet->InitWeights(1.0);
		mSearchConfig->Set("disableOptimisticEstimate", "true");
	}
}

IterativeSearch::~IterativeSearch() {
	delete mSearchConfig;
	delete mSearch;
	delete mUpdatedDescDataSet;
	delete mUpdatedModelDataSet;
}
void IterativeSearch::LogParameters() {
	Search::LogParameters();
}
string IterativeSearch::GetName() {
	if(mWeight == 0.0)
		return string("sequential cover w/ ") + mSearch->GetName();
	else {
		char temp[20]; sprintf_s(temp, 20, "%.02f", mWeight);
		return string("weighted cover (") + string(temp) + string(") w/ ") + mSearch->GetName();
	}
}
SubgroupCollection *IterativeSearch::DoSearch() {
	InitTimer();

	// Init
	uint32 numPatterns = mConfig->Read<uint32>("topK");
	SubgroupCollection *results = new SubgroupCollection(mConfig->Read<uint32>("topK"));

	// Do iterative mining
	for(uint32 i=0; i<numPatterns; i++) {
		if(mUpdatedDescDataSet->GetNumRows() <= mConfig->Read<uint32>("minCoverage"))
			break;

		Search *search = Create(mSearchDesc, mSearchConfig, mUpdatedDescDataSet, mUpdatedModelDataSet);
		SubgroupCollection *iterationResults = search->DoSearch();

		if(iterationResults->GetSize() == 0) {
			delete iterationResults;	// nothing found
			delete search;
			break;
		}

		// Get single subgroup
		Subgroup *subgroup = *(iterationResults->GetSubgroups()->begin());

		// Update (weights of) dataset
		if(mWeight == 0.0)
			SequentialCover(subgroup);
		else
			WeightedCover(subgroup);
		
		// Translate subgroup subset to full database
		ConditionArray *ca = subgroup->GetConditions();
		Subgroup *sg = new Subgroup(ca->Clone(), ca->Apply(mDescDataSet));
		mMeasure->ComputeQuality(sg);

		// Cleanup
		delete iterationResults;
		delete search;

		// Merge found pattern into overall collection
		results->Add(sg);
		delete sg; // collection makes a copy

		// Report to the user
		if(OutOfTime())
			LOG_MSG("** Out of time! Stopping right here, right now. **");
		else
			printf("Finished iteration %u.\n", i+1);

		// Perform intermediate save
		DoIntermediateSave(i, results);
	}

	return results;
}
void IterativeSearch::SequentialCover(Subgroup *subgroup) {
	mUpdatedModelDataSet->RemoveSubset(subgroup->GetSubset());	// it may not seem like it, but this construction is a hack:
																// subgroup gets the dataset size from its description dataset,
																// so changing the order here would get you into trouble.
	mUpdatedDescDataSet->RemoveSubset(subgroup->GetSubset());
}
void IterativeSearch::WeightedCover(Subgroup *subgroup) {
	mUpdatedModelDataSet->MultiplyWeights(subgroup->GetSubset(), mWeight);
}
