//
// Copyright (c) 2011-2012, Matthijs van Leeuwen
//
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
// Neither the name of the Universiteit Utrecht, KU Leuven, nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
#include "../elobal.h"

#if defined (__unix__)
#include <glibc_s.h>
#endif

#include <Config.h>
#include <ArrayUtils.h>
#include "../datastructs/conditions/ConditionArray.h"
#include "../datastructs/dataset/DataSet.h"
#include "../datastructs/data/BinaryColumn.h"
#include "../datastructs/data/FloatColumn.h"
#include "../datastructs/data/NominalColumn.h"
#include "../datastructs/subgroup/Subgroup.h"
#include "CandidateQueue.h"

#include "RefinementOperator.h"

RefinementOperator::RefinementOperator(Config *config, DataSet *ds) : mDataSet(ds) {
	mColumns = mDataSet->GetColumns();
	mNumCols = mDataSet->GetNumColumns();
	mColIds = mDataSet->GetColumnIds();

	mFloatNumSplits = config->Read<uint32>("floatNumSplits");

	mTempFloats = new float[ds->GetNumRows()];
}

RefinementOperator::~RefinementOperator() {
	delete[] mColIds;
	delete[] mTempFloats;
}
void RefinementOperator::Refine(Subgroup *subgroup, CandidateQueue *candQueue) {
	// Foreach column
	for(uint32 i=0; i<mNumCols; i++)
		RefineColumn(mColumns[i], subgroup, candQueue);
}
void RefinementOperator::RefineColumn(Column *column, Subgroup *subgroup, CandidateQueue *candQueue) {
	// Refine according to column type
	switch(column->GetType()) {
	case BinaryColumnType:
		RefineBinary((BinaryColumn *)column, subgroup, candQueue);
		break;
	case FloatColumnType:
		RefineFloat((FloatColumn *)column, subgroup, candQueue);
		break;
	case NominalColumnType:
		RefineNominal((NominalColumn *)column, subgroup, candQueue);
		break;
	default:
		THROW("Unsupported column type?!");
	}
}
void RefinementOperator::RefineBinary(BinaryColumn *column, Subgroup *subgroup, CandidateQueue *candQueue) {
	// Refining only makes sense if not yet used
	if(subgroup->GetConditions()->IsColumnUsed(column->GetId()))
		return;

	// == true
	candQueue->Add(new ConditionArray(subgroup, column->GetId(), Equal, true));
	// == false
	candQueue->Add(new ConditionArray(subgroup, column->GetId(), Equal, false));
}
void RefinementOperator::RefineFloat(FloatColumn *column, Subgroup *subgroup, CandidateQueue *candQueue) {
	// Float always refined in current implementation, meaning that ranges can be gradually refined, slowly peeling off data. This may cause redundant conditions in the outcome.
	// (However, note that these redundant conditions are automatically removed when dominance based pruning is applied.)

	// Get the required info
	uint32 numValues = subgroup->GetSize();
	column->GetMaskedFloatValuesIn(subgroup->GetMask(), mTempFloats);

	// Create histogram online, for `local' splitting. [Could be cached if doing on a global scale]
	ArrayUtils::CreateStaticMess<float>(numValues, false);
	ArrayUtils::MSortArrayAsc<float>(mTempFloats, numValues);
	ArrayUtils::CleanUpStaticMess();

	// Equal bin-size splitting
	uint32 numSplits, delta;
	if(numValues <= mFloatNumSplits) { // cannot do more splits than numValues-1
		numSplits = numValues-1; 
		delta = 1;
	} else {
		numSplits = mFloatNumSplits;
		delta = numValues / mFloatNumSplits;
	}

	// Do the splitting
	for(uint32 i=1; i<=numSplits; i++) { // foreach split point
		float val = (mTempFloats[i * delta - 1] + mTempFloats[i * delta]) / 2;
		// < val
		candQueue->Add(new ConditionArray(subgroup, column->GetId(), LowerThan, val));
		// > val
		candQueue->Add(new ConditionArray(subgroup, column->GetId(), GreaterThan, val));
	}
}
void RefinementOperator::RefineNominal(NominalColumn *column, Subgroup *subgroup, CandidateQueue *candQueue) {
	ConditionArray *conditions = subgroup->GetConditions();

	if(conditions->IsColumnUsed(column->GetId())) {
		// Refining does not make sense if a single value is already specified
		if(conditions->ContainsEqualsForColumn(column->GetId()))
			return;

		// Column is used, but with != condition(s). Allow new != conditions on this column, but not those that are already there.
		uint32 numValues = column->GetNumValues();
		for(uint32 i=0; i<numValues; i++) {
			if(!conditions->Contains(column->GetId(), NotEqual, i))
				candQueue->Add(new ConditionArray(subgroup, column->GetId(), NotEqual, i));
		}

	} else {	// Column not yet used
		// Assume > 2 values
		uint32 numValues = column->GetNumValues();
		for(uint32 i=0; i<numValues; i++) {
			// == val
			candQueue->Add(new ConditionArray(subgroup, column->GetId(), Equal, i));
			// != val
			candQueue->Add(new ConditionArray(subgroup, column->GetId(), NotEqual, i));
		}
	}
}
void RefinementOperator::RefineOrdered(Subgroup *subgroup, CandidateQueue *candQueue) {
	// Seek to last currently used column
	uint32 startFrom = 0;
	if(subgroup->GetConditions()->GetSize() > 0) {
		uint32 lastCol = subgroup->GetConditions()->GetLastColumnId();
		for(; startFrom<mNumCols; startFrom++)
			if(mColIds[startFrom] == lastCol)
				break;
	}

	// Skip to first unused column (disable this to allow multiple conditions per attribute, but then use RefineColumn instead of RefineColumnNoChecks!)
	//startFrom++;
	
	// Start from there
	for(uint32 i=startFrom; i<mNumCols; i++)
		RefineColumn(mColumns[i], subgroup, candQueue); // ..NoChecks
}
void RefinementOperator::RefineColumnNoChecks(Column *column, Subgroup *subgroup, CandidateQueue *candQueue) {
	// Refine according to column type
	switch(column->GetType()) {
		case BinaryColumnType:
			RefineBinaryNoChecks((BinaryColumn *)column, subgroup, candQueue);
			break;
		case FloatColumnType:
			RefineFloatNoChecks((FloatColumn *)column, subgroup, candQueue);
			break;
		case NominalColumnType:
			RefineNominalNoChecks((NominalColumn *)column, subgroup, candQueue);
			break;
		default:
			THROW("Unsupported column type?!");
	}
}
void RefinementOperator::RefineBinaryNoChecks(BinaryColumn *column, Subgroup *subgroup, CandidateQueue *candQueue) {
	// == true
	candQueue->Add(new ConditionArray(subgroup, column->GetId(), Equal, true));
	// == false
	candQueue->Add(new ConditionArray(subgroup, column->GetId(), Equal, false));
}
void RefinementOperator::RefineFloatNoChecks(FloatColumn *column, Subgroup *subgroup, CandidateQueue *candQueue) {
	// Get the required info
	uint32 numValues = subgroup->GetSize();
	column->GetMaskedFloatValuesIn(subgroup->GetMask(), mTempFloats);

	// Create histogram online, for `local' splitting. [Could be cached if doing on a global scale]
	ArrayUtils::CreateStaticMess<float>(numValues, false);
	ArrayUtils::MSortArrayAsc<float>(mTempFloats, numValues);
	ArrayUtils::CleanUpStaticMess();

	// Equal bin-size splitting
	uint32 numSplits, delta;
	if(numValues <= mFloatNumSplits) { // cannot do more splits than numValues-1
		numSplits = numValues-1; 
		delta = 1;
	} else {
		numSplits = mFloatNumSplits;
		delta = numValues / mFloatNumSplits;
	}

	// Do the splitting
	for(uint32 i=1; i<=numSplits; i++) { // foreach split point
		float val = (mTempFloats[i * delta - 1] + mTempFloats[i * delta]) / 2;
		// < val
		candQueue->Add(new ConditionArray(subgroup, column->GetId(), LowerThan, val));
		// > val
		candQueue->Add(new ConditionArray(subgroup, column->GetId(), GreaterThan, val));
	}
}
void RefinementOperator::RefineNominalNoChecks(NominalColumn *column, Subgroup *subgroup, CandidateQueue *candQueue) {
	// Assume > 2 values
	uint32 numValues = column->GetNumValues();
	for(uint32 i=0; i<numValues; i++) {
		// == val
		candQueue->Add(new ConditionArray(subgroup, column->GetId(), Equal, i));
		// != val
		candQueue->Add(new ConditionArray(subgroup, column->GetId(), NotEqual, i));
	}
}
RefinementOperator *RefinementOperator::Create(Config *config, DataSet *ds) {
	return new RefinementOperator(config, ds);
}
