//
// Copyright (c) 2011-2012, Matthijs van Leeuwen
//
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
// Neither the name of the Universiteit Utrecht, KU Leuven, nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
#include "../elobal.h"

#if defined (__unix__)
#include <glibc_s.h>
#endif

#include <Config.h>
#include <TimeUtils.h>
#include <logger/Log.h>

#include "../measures/QualityMeasure.h"
#include "RefinementOperator.h"
#include "IterativeSearch.h"
#include "beam/BeamSearch.h"
#include "bfs/BreadFirstSearch.h"
#include "dfs/DepthFirstSearch.h"

#include "Search.h"

Search::Search(Config *config, DataSet *descDS, DataSet *modelDS) {
	mConfig = config;
	mDescDataSet = descDS;
	mModelDataSet = modelDS;

	mRefineOp = RefinementOperator::Create(config, descDS);
	mMeasure = QualityMeasure::Create(config, modelDS);

	mMinCoverage	= config->Read<uint32>("minCoverage");
	mMaxDepth		= config->Read<uint32>("maxDepth");
	mDisableOptimisticEstimate = config->Read<bool>("disableOptimisticEstimate", false);
	mUseOptimisticEstimate = !mDisableOptimisticEstimate && mMeasure->HasOptimisticEstimate();

	mIntermediateSave = config->Read<bool>("outputIntermediate");
	mIntermediateSaveDir = Bass::GetWorkingDir() + "intermediate/";
	mMaxTime = config->Read<uint32>("maxTime", 0) * 60; // from minutes to seconds
	mTimer = NULL;
}

Search::~Search() {
	delete mRefineOp;
	delete mMeasure;
	delete mTimer;
}
void Search::LogParameters() {
	LOG_MSG("Max time:\t\t" + mConfig->Read<string>("maxTime"));
	LOG_MSG("Search type:\t\t" + mConfig->Read<string>("searchType"));
	LOG_MSG("Min coverage:\t\t" + mConfig->Read<string>("minCoverage"));
	LOG_MSG("Max depth:\t\t" + mConfig->Read<string>("maxDepth"));
	LOG_MSG("Float # splits:\t\t" + mConfig->Read<string>("floatNumSplits"));
	LOG_MSG("Top-k:\t\t\t" + mConfig->Read<string>("topK"));
	LOG_MSG(string("Using OE:\t\t") + (CanUseOptimisticEstimate() && mUseOptimisticEstimate ? "yes" : "no"));
	mMeasure->LogParameters();
}
Search* Search::Create(Config *config, DataSet *descDS, DataSet *modelDS) {
	string desc = config->Read<string>("searchType");
	return Create(desc, config, descDS, modelDS);
}
Search* Search::Create(const string &desc, Config *config, DataSet *descDS, DataSet *modelDS) {
	if(desc.compare("beam") == 0)
		return new BeamSearch(config, descDS, modelDS);
	if(desc.compare("bfs") == 0)
		return new BreadFirstSearch(config, descDS, modelDS);
	if(desc.compare("dfs") == 0)
		return new DepthFirstSearch(config, descDS, modelDS);
	if(desc.substr(0, 5).compare("iter-") == 0) {
		double weight = atof(desc.substr(5,3).c_str());
		return new IterativeSearch((float)weight, desc.substr(9), config, descDS, modelDS);
	}
	THROW("Unknown measure type.");
}
void Search::DoIntermediateSave(const uint32 depth, SubgroupCollection *results) {
	if(!HasIntermediateSave() || !mIntermediateSave || depth==mMaxDepth) // no intermediate saving OR final level anyway
		return;
	FileUtils::CreateDirIfNotExists(mIntermediateSaveDir);
	char tmp[5]; sprintf_s(tmp, 5, "%02u", depth);
	results->WriteStatsFile(mIntermediateSaveDir + "level" + tmp + "-" + Logger::GetStartTimeString() + ".csv", mDescDataSet);
}
uint32 Search::GetElapsedTime() {
	return (uint32)(mTimer->GetElapsedTime() / 60); // sec to min
}
void Search::InitTimer() {
	delete mTimer;
	mTimer = new Timer();	
}
bool Search::OutOfTime() {
	if(mMaxTime == 0)
		return false;
	return mTimer->GetElapsedTime() > mMaxTime;
}
