//
// Copyright (c) 2011-2012, Matthijs van Leeuwen
//
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
// Neither the name of the Universiteit Utrecht, KU Leuven, nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
#include "../../elobal.h"

#if defined (__unix__)
#include <glibc_s.h>
#endif

#include <Config.h>
#include <logger/Log.h>

#include "../../datastructs/conditions/ConditionArray.h"
#include "../../datastructs/dataset/DataSet.h"
#include "../../datastructs/datasubset/DataSubset.h"
#include "../../datastructs/subgroup/Subgroup.h"
#include "../../measures/QualityMeasure.h"
#include "../CandidateQueue.h"
#include "../RefinementOperator.h"
#include "BeamSelector.h"

#include "BeamSearch.h"

BeamSearch::BeamSearch(Config *config, DataSet *descDS, DataSet *modelDS) : Search(config, descDS, modelDS) {
	// Configuration
	mBeamSelector = BeamSelector::Create(config, descDS, modelDS);
	mBeamSelectFromTopK = config->Read<bool>("beamSelectFromTopK", false);	// deprecated
	mBeamPostSelect = config->Read<bool>("beamPostSelect", false);			// deprecated
	mLogBeamSelection = config->Read<bool>("beamLogSelection", false);
}
BeamSearch::~BeamSearch() {
	// not my mDescDS, mModelDS
	delete mBeamSelector;
}
void BeamSearch::LogParameters() {
	Search::LogParameters();
	mBeamSelector->LogParameters();
//	LOG_MSG("Beam select from top-k:\t" + string(mBeamSelectFromTopK?"yes":"no"));
//	LOG_MSG("Beam post selection:\t" + string(mBeamPostSelect?"yes":"no"));
}
SubgroupCollection	*BeamSearch::DoSearch() {
	char temp[100];
	InitTimer();

	// Init some structures
	SubgroupCollection *results = new SubgroupCollection(mConfig->Read<uint32>("topK"), mConfig->Read<string>("sgSubsets","cache").compare("rebuild")!=0);
	CandidateQueue *candidates = new CandidateQueue();
	SubgroupList *subgroups = new SubgroupList();
	SubgroupList *beam = new SubgroupList();

	// Init beam log
	FILE *mBeamLogFP = NULL;
	if(mLogBeamSelection) {
		string s = Bass::GetWorkingDir() + "beamselection.log";
		fopen_s(&mBeamLogFP, s.c_str(), "w");
	}

	// Init beam with empty root candidate
	{
		uint32 numRows = mDescDataSet->GetNumRows();
		bool *mask = new bool[numRows];
		for(uint32 i=0; i<numRows; i++)
			mask[i] = true;
		Subgroup *root = new Subgroup(new ConditionArray(), DataSubset::Create(mDescDataSet, mask, true));
		beam->push_back(root);
	}

	// While candidates and within maxDepth
	for(uint32 depth=1; depth <= mMaxDepth && !OutOfTime(); ++depth) {
		// Generate next level candidates
		printf_s("Generating level-%u candidates ...\r", depth);
		for(SubgroupList::iterator it=beam->begin(); it!=beam->end(); ++it) {
			(*it)->CacheSubset(mDescDataSet);
			mRefineOp->Refine(*it, candidates);
		}
		uint32 numCands = candidates->GetSize();
		printf_s("Generated %u level-%u candidates.\n", numCands, depth);
		
		// Score candidates
		uint32 curCand = 0;
		printf_s("Testing level-%u candidates ... \r", depth);
		while(!candidates->Empty() && !OutOfTime()) {
			// Create  associated subset
			ConditionArray *cand = candidates->GetNext();
			DataSubset *subset = cand->Apply(mDescDataSet, cand->GetParent()->GetMask());

			if(subset->GetSize() >= mMinCoverage && subset->GetSize() < mDescDataSet->GetNumRows()) {
				// Compute quality
				Subgroup *sg = new Subgroup(cand, subset);
				double quality = mMeasure->ComputeWeightedQuality(sg);

				// Add to result collection
				results->Add(sg);

				// Add to current level using insertion sort (note: could be optimized, but this leaves all possibilities for beam selection)
				SubgroupList::iterator it = subgroups->begin();
				while(it!=subgroups->end() && (*it)->GetQuality() >= quality)
					++it;
				subgroups->insert(it, sg);

			} else {	// discard
				delete cand;
				delete subset;
			}
			++curCand;
		}
		results->AddToCandidateCount(curCand);

		// Cleanup if out of time
		while(!candidates->Empty())	
			delete candidates->GetNext();

		// Report to the user
		if(OutOfTime())
			LOG_MSG("** Out of time! Stopping right here, right now. **");
		else
			printf("Finished testing level-%u candidates.\n", depth);

		// Perform intermediate save
		DoIntermediateSave(depth, results);

		// Select new beam
		Cleanup(beam);
		if(depth < mMaxDepth && !OutOfTime() && subgroups->size()>0) {
			uint32 numBeamCandidates;

			// Assign subgroup ids if logging enabled
			if(mLogBeamSelection) { 
				uint32 id = 0;
				for(SubgroupList::iterator it=subgroups->begin(); it!=subgroups->end(); ++it)
					(*it)->SetId(id++);
			}

			// Do actual beam selection
			if(mBeamSelectFromTopK) {
				Cleanup(subgroups); // results from current level should be deleted
				results->GetUnrefinedSubgroupsIn(subgroups);
				numBeamCandidates = (uint32) subgroups->size();
				mBeamSelector->Select(subgroups, beam, depth, false);					// select beam from top-k results so far, don't take ownership
			} else {
				numBeamCandidates = (uint32) subgroups->size();
				mBeamSelector->Select(subgroups, beam, depth, true);					// select beam from results of last level, taking ownership
			}
			subgroups->clear();

			// Log beam selection if enabled
			if(mLogBeamSelection) { // log if required 
				fprintf_s(mBeamLogFP, "%u:", numBeamCandidates);
				for(SubgroupList::iterator it=beam->begin(); it!=beam->end(); ++it)
					fprintf_s(mBeamLogFP, "%u,", (*it)->GetId()); // (unsorted id list!)
				fprintf_s(mBeamLogFP, "\n");
			}

			sprintf_s(temp, 100, "Included %u subgroups in beam.", beam->size());
			LOG_MSG(string(temp));
		} else {
			Cleanup(subgroups);
		}
	}

	// Do BeamSelector post-processing if asked for (deprecated)
	if(mBeamPostSelect) {
		LOG_MSG(" * Doing beam selection on final results.");
		SubgroupList *subgroups = results->GetSubgroups();
		mBeamSelector->Select(subgroups, beam, mMaxDepth, true);
		subgroups->clear();
		for(SubgroupList::iterator it=beam->begin(); it!=beam->end(); it++)
			subgroups->push_back(*it);
		beam->clear();
		results->UpdateStats();
	}

	// Cleanup
	delete beam;
	delete subgroups;
	delete candidates;
	if(mLogBeamSelection)
		fclose(mBeamLogFP);

	// Return resulting subgroups
	return results;
}
void BeamSearch::Cleanup(SubgroupList *sgList) {
	for(SubgroupList::iterator it=sgList->begin(); it!=sgList->end(); ++it)
		delete *it;
	sgList->clear();
}
