//
// Copyright (c) 2011-2012, Matthijs van Leeuwen
//
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
// Neither the name of the Universiteit Utrecht, KU Leuven, nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
#include "../../elobal.h"

#if defined (__unix__)
#include <glibc_s.h>
#endif

#include <Bass.h>
#include <Config.h>
#include <logger/Log.h>

#include "../../datastructs/dataset/DataSet.h"
#include "../../datastructs/subgroup/Subgroup.h"

#include "QualityBeam.h"
#include "DescriptionBeam.h"
#include "CoverBeam.h"
#include "CompressionBeam.h"

#include "BeamSelector.h"

BeamSelector::BeamSelector(Config *config, DataSet *descDS, DataSet *modelDS) {
	mDescDS = descDS;
	mModelDS = modelDS;
	mNumRows = modelDS->GetNumRows();

	mMaxBeamWidth = config->Read<uint32>("beamWidth");
	string var = config->Read<string>("beamVarWidth", "false");
	mUseVariableWidth = var.compare("false")!=0;
	mVariableParam = mUseVariableWidth ? config->Read<float>("beamVarWidth") : 0.0f;
}

BeamSelector::~BeamSelector() {

}
void BeamSelector::LogParameters() {
	LOG_MSG("Beam type:\t\t" + GetName());
	char temp[20]; sprintf_s(temp, 20, "%u", mMaxBeamWidth);
	LOG_MSG("Max beam width:\t\t" + string(temp));
	LOG_MSG(string("Variable beam width:\t") + (mUseVariableWidth?"yes":"no"));
	if(mUseVariableWidth) {
		sprintf_s(temp, 20, "%.02f", mVariableParam);
		LOG_MSG("Variable beam param:\t" + string(temp));
	}
}
string BeamSelector::GetConfigString() {
	char temp[100];
	if(mUseVariableWidth)
		sprintf_s(temp, 100, "%u-var-%.02f", mMaxBeamWidth, mVariableParam);
	else
		sprintf_s(temp, 100, "%u-fixed", mMaxBeamWidth);
	return string(temp);
}
void BeamSelector::AddToBeam(SubgroupList *beam, Subgroup *subgroup, const bool takeOwnership) {
	subgroup->SetRefined(true);
	beam->push_back(takeOwnership ? subgroup : subgroup->Clone()); 
}
BeamSelector *BeamSelector::Create(Config *config, DataSet *descDS, DataSet *modelDS) {
	string desc = config->Read<string>("beamStrategy");
	if(desc.compare("quality") == 0)						// Standard top-k selection based on quality only
		return new QualityBeam(config, descDS, modelDS);
	if(desc.compare("description") == 0)					// Description-based selection
		return new DescriptionBeam(config, descDS, modelDS);
	if(desc.compare("cover") == 0)							// Cover-based selection
		return new CoverBeam(config, descDS, modelDS);
	if(desc.compare("compression") == 0)					// Compression-based selection
		return new CompressionBeam(config, descDS, modelDS);
	THROW("Unknown beam selector type.");
}
