//
// Copyright (c) 2011-2012, Matthijs van Leeuwen
//
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
// Neither the name of the Universiteit Utrecht, KU Leuven, nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
#include "../../elobal.h"

#if defined (__unix__)
#include <glibc_s.h>
#endif

#include <Config.h>
#include <logger/Log.h>

#include "../../datastructs/dataset/DataSet.h"
#include "../../datastructs/subgroup/Subgroup.h"
#include "../../models/compression/Compressor.h"

#include "CompressionBeam.h"

CompressionBeam::CompressionBeam(Config *config, DataSet *descDS, DataSet *modelDS) : BeamSelector(config, descDS, modelDS) {
	// Init
	mBaseSizes = new float[mNumRows];
	mCurSizes = new float[mNumRows];
	mTestSizes = new float[mNumRows];

	// Compute base compressed sizes
	if(!modelDS->GetModel()->IsCompressor())
		THROW("Can only use compression-based models.");
	Compressor *baseCompressor = (Compressor *)(modelDS->GetModel());
	baseCompressor->GetEncodedTransactionSizes(modelDS, mBaseSizes);

	// Check first whether models are cached
	if(config->Read<string>("sgModels").compare("cache") != 0)
		THROW("We've got a problem. Are you sure models are cached? (sgModels = cache)");
}
CompressionBeam::~CompressionBeam() {
	delete[] mBaseSizes;
	delete[] mCurSizes;
	delete[] mTestSizes;
}
string CompressionBeam::GetName() {
	if(mUseVariableWidth) {
		return string("compression-var");
	}
	return string("compression");
}
void CompressionBeam::LogParameters() {
	BeamSelector::LogParameters();
}
void CompressionBeam::Select(SubgroupList *pool, SubgroupList *beam, const uint32 depth, const bool takeOwnership) {
	// Init the list that we are going to modify
	SubgroupList *subgroups;
	if(takeOwnership) {
		// SubgroupList pool is going away anyhow, simply use that one.
		subgroups = pool;
	} else {
		// We are not allowed to touch the pool, but since we want to modify the list, create a copy here
		subgroups = new SubgroupList();
		subgroups->assign(pool->begin(), pool->end());
	}

	// Init best sizes with base sizes
	memcpy(mCurSizes, mBaseSizes, mNumRows * sizeof(float));

	// Each round, select that subgroup that improves compression the most
	for(uint32 numSelected=0; numSelected<mMaxBeamWidth && !subgroups->empty(); numSelected++) {
		// Init with first
		SubgroupList::iterator selIter = subgroups->begin();
		double selGain = ComputeGain(*selIter);

		// Test rest
		SubgroupList::iterator it = subgroups->begin(); ++it; // skip first
		for(; it!=subgroups->end(); ++it) {
			double gain = ComputeGain(*it);
			if(gain > selGain) { // Better! Select this one instead
				selIter = it;
				selGain = gain;
			}
		}

		// Variable beam size? (this makes beam size the maximum width)
		if(mUseVariableWidth && selGain <= mVariableParam)
			break;	// stop immediately if best gain is lower than the specified minimum

		// Add to beam
		AddToBeam(beam, *selIter, takeOwnership);
		// Update counts used to compute overlap
		Compressor *compressor = (Compressor *)((*selIter)->GetModel());
		compressor->GetEncodedTransactionSizes(mModelDS, mCurSizes, (*selIter)->GetSubset());
		/*DataSubset *subset = (*selIter)->GetSubset(); // -- Use shortest encoded length instead of last
		compressor->GetEncodedTransactionSizes(mModelDS, mCurSizes, subset);
		for(uint32 r=0; r<mNumRows; r++)
			if(subset->IsInSubset(r) && mTestSizes[r] < mCurSizes[r])
				mCurSizes[r] = mTestSizes[r];*/
		// Remove from candidate list
		subgroups->erase(selIter);
	}

	if(takeOwnership) {
		// Delete all items that have not been selected for the beam
		for(SubgroupList::iterator it=subgroups->begin(); it!=subgroups->end(); ++it)
			delete *it;
	} else {
		// Keep everything in pool untouched, but temporary subgroups list should be deleted
		delete subgroups;
	}
}
float CompressionBeam::ComputeGain(Subgroup *subgroup) {
	Compressor *compressor = (Compressor *)(subgroup->GetModel());
	DataSubset *subset = subgroup->GetSubset();
	compressor->GetEncodedTransactionSizes(mModelDS, mTestSizes, subset);
	float gain = 0.0f;
	for(uint32 r=0; r<mNumRows; r++)
		if(subset->IsInSubset(r))
			gain += mCurSizes[r] - mTestSizes[r];
	return gain;
}
