//
// Copyright (c) 2011-2012, Matthijs van Leeuwen
//
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
// Neither the name of the Universiteit Utrecht, KU Leuven, nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
#include "../../elobal.h"

#if defined (__unix__)
#include <glibc_s.h>
#endif

#include <Config.h>
#include <logger/Log.h>

#include "../../datastructs/subgroup/Subgroup.h"

#include "CoverBeam.h"

CoverBeam::CoverBeam(Config *config, DataSet *descDS, DataSet *modelDS) : BeamSelector(config, descDS, modelDS) {
	// Config
	mMultiplicativeWeight = config->Read<float>("coverBeamMultiplier");
	string coverDesc = config->Read<string>("coverBeamStrategy");
	if(coverDesc.compare("multiplicative") == 0)
		mOverlapFunc = &CoverBeam::MultiplicativeWeightCover;
	else if(coverDesc.compare("additive") == 0)
		mOverlapFunc = &CoverBeam::AdditiveWeightCover;
	else
		mOverlapFunc = &CoverBeam::SequentialCover;
}
CoverBeam::~CoverBeam() {

}
string CoverBeam::GetName() {
	if(mUseVariableWidth) {
		char temp[100];
		sprintf_s(temp, 100, "cover-var(%.1f)", mVariableParam);
		return string(temp);
	}
	return string("cover");
}
void CoverBeam::LogParameters() {
	BeamSelector::LogParameters();
	if(mOverlapFunc == &CoverBeam::MultiplicativeWeightCover) {
		LOG_MSG("Weight strategy:\tmultiplicative weighted covering"); 
		char temp[20]; sprintf_s(temp, 20, "%.03f", mMultiplicativeWeight);
		LOG_MSG("Weight multiplier:\t"+string(temp));
	} else if(mOverlapFunc == &CoverBeam::AdditiveWeightCover)
		LOG_MSG("Weight strategy:\tadditive weighted covering");
	else
		LOG_MSG("Weight strategy:\tsequential covering");
}
void CoverBeam::Select(SubgroupList *pool, SubgroupList *beam, const uint32 depth, const bool takeOwnership) {
	SubgroupList *subgroups;
	if(takeOwnership) {
		// SubgroupList pool is going away anyhow, simply use that one.
		subgroups = pool;
	} else {
		// We are not allowed to touch the pool, but since we want to modify the list, create a copy here
		subgroups = new SubgroupList();
		subgroups->assign(pool->begin(), pool->end());
	}

	// Always select the highest ranking subgroup
	AddToBeam(beam, *(subgroups->begin()), takeOwnership);
	double minScore = mUseVariableWidth ? 
						mVariableParam * (*(subgroups->begin()))->GetQuality() : // * 1.0 for coverageScore
						0.0; // score cannot fall below 0

	// Remove from candidate list
	subgroups->pop_front();
	// Init counts for overlap computation
	uint32 *counts = InitCounts(*(beam->begin()));

	// For each additional beam element, select that one that has minimal overlap with maximum score
	for(uint32 numSelected=1; numSelected<mMaxBeamWidth && !subgroups->empty(); numSelected++) {
		float coverageScore;

		// Use first (highest ranking) by default
		SubgroupList::iterator selIter = subgroups->begin();
		coverageScore = (this->*(mOverlapFunc))(*selIter, counts);
		double selScore = (*selIter)->GetQuality() * coverageScore;

		// Test rest
		SubgroupList::iterator it = subgroups->begin(); ++it; // skip first
		for(; it!=subgroups->end(); ++it) {
			coverageScore = (this->*(mOverlapFunc))(*it, counts);
			double score = (*it)->GetQuality() * coverageScore;
			if(score > selScore) { // Better! Select this one instead
				selIter = it;
				selScore = score;
			}
		}

		// Variable beam width: stop when highest score falls below specified (relative) minimum
		if(selScore < minScore)
			break;

		// Add to beam
		AddToBeam(beam, *selIter, takeOwnership);
		// Update counts used to compute overlap
		UpdateCounts(*selIter, counts);
		// Remove from candidate list
		subgroups->erase(selIter);
	}

	if(takeOwnership) {
		// Delete all items that have not been selected for the beam
		for(SubgroupList::iterator it=subgroups->begin(); it!=subgroups->end(); ++it)
			delete *it;
	} else {
		// Keep everything in pool untouched, but temporary subgroups list should be deleted
		delete subgroups;
	}

	// And cleanup the counts
	delete[] counts;
}
uint32 *CoverBeam::InitCounts(Subgroup *subgroup) {
	uint32 *counts = new uint32[mNumRows];
	bool *mask = subgroup->GetMask();
	for(uint32 i=0; i<mNumRows; i++)
		counts[i] = mask[i] ? 1 : 0;
	return counts;
}
void CoverBeam::UpdateCounts(Subgroup *subgroup, uint32 *counts) {
	bool *mask = subgroup->GetMask();
	for(uint32 i=0; i<mNumRows; i++)
		if(mask[i])
			++counts[i];
}
float CoverBeam::SequentialCover(Subgroup *subgroup, uint32 *counts) {
	float score = 0.0f;
	bool *mask = subgroup->GetMask();
	for(uint32 i=0; i<mNumRows; i++)
		if(mask[i] && counts[i]==0)
			score += 1.0f;
	score /= subgroup->GetSize();
	return score;						// simply returns the percentage of subgroup that does NOT overlap with any part of the database already covered
}
float CoverBeam::MultiplicativeWeightCover(Subgroup *subgroup, uint32 *counts) {
	float score = 0.0f;
	bool *mask = subgroup->GetMask();
	for(uint32 i=0; i<mNumRows; i++)
		if(mask[i]) {
			if(counts[i]==0)
				score += 1.0f;
			else
				score += pow(mMultiplicativeWeight, (float)counts[i]);
		}
	score /= subgroup->GetSize();
	return score;
}
float CoverBeam::AdditiveWeightCover(Subgroup *subgroup, uint32 *counts) {
	float score = 0.0f;
	bool *mask = subgroup->GetMask();
	for(uint32 i=0; i<mNumRows; i++)
		if(mask[i]) {
			if(counts[i]==0)
				score += 1.0f;
			else
				score += 1.0f/(1.0f+counts[i]);
		}
	score /= subgroup->GetSize();
	return score;
}
