//
// Copyright (c) 2011-2012, Matthijs van Leeuwen
//
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
// Neither the name of the Universiteit Utrecht, KU Leuven, nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
#include "../../elobal.h"

#if defined (__unix__)
#include <glibc_s.h>
#endif

#include <Config.h>
#include <logger/Log.h>

#include "../../datastructs/conditions/ConditionArray.h"
#include "../../datastructs/dataset/DataSet.h"
#include "../../datastructs/subgroup/Subgroup.h"

#include "DescriptionBeam.h"

DescriptionBeam::DescriptionBeam(Config *config, DataSet *descDS, DataSet *modelDS) : BeamSelector(config, descDS, modelDS) {
	// Config
	mMinDifference = config->Read<uint32>("descriptionMinDiff", 2); // default to 2
}
DescriptionBeam::~DescriptionBeam() {

}
void DescriptionBeam::LogParameters() {
	BeamSelector::LogParameters();
}
string DescriptionBeam::GetName() {
	if(mUseVariableWidth) {
		char temp[100];
		sprintf_s(temp, 100, "description-var(%d)", (uint32)floor(mVariableParam));
		return string(temp);
	}
	return string("description");
}
void DescriptionBeam::Select(SubgroupList *subgroups, SubgroupList *beam, const uint32 depth, const bool takeOwnership) {
	SubgroupList::iterator it = subgroups->begin();

	if(mUseVariableWidth) {							// Variable beam width: each condition may occur at most c*depth times in beam
		// Init
		uint32 maxUsage = (uint32)floor(mVariableParam) * (*it)->GetConditions()->GetSize(); // each column allowed to be used (param * curDepth) times
		uint32 numColsMaxReached = 0;
		uint32 numCols = mDescDS->GetNumColumns();
		uint16 *columnUsages = new uint16[numCols];
		for(uint32 c=0; c<numCols; c++)
			columnUsages[c] = 0;

		// Count usages of first beam element
		ConditionArray *conds = (*it)->GetConditions();
		for(uint32 i=0; i<conds->GetSize(); i++)
			columnUsages[conds->GetColumnId(i)]++;

		// Always select the highest ranking subgroup
		AddToBeam(beam, *it, takeOwnership);
		++it;

		// For each additional subgroup, only select if it unequal (sufficiently different) from all subgroups already selected
		for(uint32 numSelected=1; numSelected<mMaxBeamWidth && numColsMaxReached<numCols && it!=subgroups->end(); it++) {
			bool selectThisOne = true;

			// Check whether any of the conditions has reached its usage limit
			ConditionArray *conds = (*it)->GetConditions();
			for(uint32 i=0; i<conds->GetSize(); i++) {
				if(columnUsages[conds->GetColumnId(i)] >= maxUsage) {
					selectThisOne = false;
					break;
				}
			}

			if(selectThisOne) {
				// Count usages of first beam element
				for(uint32 i=0; i<conds->GetSize(); i++) {
					columnUsages[conds->GetColumnId(i)]++;
					if(columnUsages[conds->GetColumnId(i)] == maxUsage)
						numColsMaxReached++;
				}

				// Add to beam
				AddToBeam(beam, *it, takeOwnership);
				numSelected++;
			} else if(takeOwnership) {
				delete *it;
			}
		}

		// Cleanup
		delete[] columnUsages;

	} else {										// Fixed beam width (original description-based selection heuristic)
		// Always select the highest ranking subgroup
		AddToBeam(beam, *it, takeOwnership);
		// Skip to next
		it++;

		// For each additional subgroup, only select if it unequal (sufficiently different) from all subgroups already selected
		for(uint32 numSelected=1; numSelected<mMaxBeamWidth && it!=subgroups->end(); it++) {
			bool selectThisOne = true;
			for(SubgroupList::iterator beamIt = beam->begin(); beamIt!=beam->end(); beamIt++) {
				if(((*it)->GetQuality()==(*beamIt)->GetQuality()) &&					// skip if: quality equal AND
					(*it)->GetConditions()->Difference((*beamIt)->GetConditions()) < mMinDifference) {		// not enough conditions different (note: this is always true at low levels)
						selectThisOne = false;
						break;
				}
			}

			if(selectThisOne) {
				// Add to beam
				AddToBeam(beam, *it, takeOwnership);
				numSelected++;
			} else if(takeOwnership) {
				delete *it;
			}
		}
	}

	// Delete all items that have not been selected for the beam
	if(takeOwnership)
		for(; it!=subgroups->end(); ++it)
			delete *it;
}
