//
// Copyright (c) 2011-2012, Matthijs van Leeuwen
//
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
// Neither the name of the Universiteit Utrecht, KU Leuven, nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
#include "../../elobal.h"

#if defined (__unix__)
#include <glibc_s.h>
#endif

#include <Config.h>
#include <logger/Log.h>

#include "../../datastructs/conditions/ConditionArray.h"
#include "../../datastructs/dataset/DataSet.h"
#include "../../datastructs/datasubset/DataSubset.h"
#include "../../datastructs/subgroup/Subgroup.h"
#include "../../datastructs/subgroup/SubgroupCollection.h"
#include "../../measures/QualityMeasure.h"
#include "../CandidateQueue.h"
#include "../RefinementOperator.h"

#include "BreadFirstSearch.h"

BreadFirstSearch::BreadFirstSearch(Config *config, DataSet *descDS, DataSet *modelDS) : Search(config, descDS, modelDS) {

}
BreadFirstSearch::~BreadFirstSearch() {
	// not my mDescDS, mModelDS
}
SubgroupCollection *BreadFirstSearch::DoSearch() {
	InitTimer();

	// Init some structures
	SubgroupCollection *results = new SubgroupCollection(mConfig->Read<uint32>("topK"), mConfig->Read<string>("sgSubsets","cache").compare("rebuild")!=0);
	CandidateQueue *candidates = new CandidateQueue();
	SubgroupList **subgroups = new SubgroupList *[mMaxDepth+1]; // do not count (empty) root node as separate level
	for(uint32 i=0; i<=mMaxDepth; i++)
		subgroups[i] = new SubgroupList();

	// Init level-1 with empty root candidate
	{
		uint32 numRows = mDescDataSet->GetNumRows();
		bool *mask = new bool[numRows];
		for(uint32 i=0; i<numRows; i++)
			mask[i] = true;
		Subgroup *root = new Subgroup(new ConditionArray(), DataSubset::Create(mDescDataSet, mask, true));
		subgroups[0]->push_back(root);
	}

	// While candidates and within maxDepth
	for(uint32 depth=1; depth <= mMaxDepth && !OutOfTime(); ++depth) {
		// Generate next level candidates
		printf("Generating level-%u candidates ...\r", depth);
		for(SubgroupList::iterator it=subgroups[depth-1]->begin(); it!=(subgroups[depth-1]->end()); ++it)
			mRefineOp->RefineOrdered(*it, candidates);
		uint32 numCands = candidates->GetSize();
		printf("Generated %u level-%u candidates.\n", numCands, depth);

		// Score candidates
		uint32 curCand = 0;
		printf("Testing level-%u candidates ... \r", depth);
		while(!candidates->Empty() && !OutOfTime()) {
			// Create  associated subset
			ConditionArray *cand = candidates->GetNext();
			DataSubset *subset = cand->Apply(mDescDataSet, cand->GetParent()->GetMask());

			if(subset->GetSize() >= mMinCoverage) {
				// Compute quality
				Subgroup *sg = new Subgroup(cand, subset);
				mMeasure->ComputeWeightedQuality(sg);
				results->Add(sg);
				subgroups[depth]->push_back(sg);

			} else {	// discard
				delete cand;
				delete subset;
			}
			++curCand;
		}
		results->AddToCandidateCount(curCand);

		// Cleanup if out of time
		while(!candidates->Empty())	
			delete candidates->GetNext();

		// Report to the user
		if(OutOfTime())
			LOG_MSG("** Out of time! Stopping right here, right now. **");
		else
			printf("Finished testing level-%u candidates.\n", depth);

		// Perform intermediate save
		DoIntermediateSave(depth, results);
	}

	// Cleanup
	for(uint32 i=0; i<=mMaxDepth; i++) {
		for(SubgroupList::iterator it = subgroups[i]->begin(); it!=subgroups[i]->end(); ++it)
			delete *it;
		delete subgroups[i];
	}
	delete[] subgroups;
	delete candidates;

	// Return resulting subgroups
	return results;
}
