//
// Copyright (c) 2011-2012, Matthijs van Leeuwen
//
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
// Neither the name of the Universiteit Utrecht, KU Leuven, nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
#include "../../elobal.h"

#if defined (__unix__)
#include <glibc_s.h>
#endif

#include <Config.h>
#include <logger/Log.h>

#include "../../datastructs/conditions/ConditionArray.h"
#include "../../datastructs/dataset/DataSet.h"
#include "../../datastructs/datasubset/DataSubset.h"
#include "../../datastructs/subgroup/Subgroup.h"
#include "../../datastructs/subgroup/SubgroupCollection.h"
#include "../../measures/QualityMeasure.h"
#include "../CandidateQueue.h"
#include "../RefinementOperator.h"

#include "DepthFirstSearch.h"

DepthFirstSearch::DepthFirstSearch(Config *config, DataSet *descDS, DataSet *modelDS) : Search(config, descDS, modelDS) {

}

DepthFirstSearch::~DepthFirstSearch() {

}
SubgroupCollection *DepthFirstSearch::DoSearch() {
	InitTimer();

	// Init some structures
	SubgroupCollection *results = new SubgroupCollection(mConfig->Read<uint32>("topK"), mConfig->Read<string>("sgSubsets","cache").compare("rebuild")!=0);

	// Create empty root candidate
	uint32 numRows = mDescDataSet->GetNumRows();
	bool *mask = new bool[numRows];
	for(uint32 i=0; i<numRows; i++)
		mask[i] = true;
	Subgroup *root = new Subgroup(new ConditionArray(), DataSubset::Create(mDescDataSet, mask, true));

	// Start from root
	printf("Searching...\r");
	Expand(1, root, results, true);
	printf("Finished search, tested %u candidates.                  \n", results->GetNumCandidates());
	if(OutOfTime())
		LOG_MSG("** Out of time! Suddenly jumped out of the tree. **");

	// Cleanup
	delete root;

	// Return resulting subgroups
	return results;
}

void DepthFirstSearch::Expand(const uint32 depth, Subgroup *node, SubgroupCollection *results, const bool showProgress /* = false */) {
	CandidateQueue *candidates = new CandidateQueue();

	// Generate candidates from node
	mRefineOp->RefineOrdered(node, candidates);

	// Score candidates
	uint32 numCands = candidates->GetSize();
	uint32 curCand = 1;
	while(!candidates->Empty()) {
		// Display progress if asked for (usually only in root)
		if(showProgress)
			printf_s("DFS: Currently exploring branch %u of %u.\r", curCand, numCands);

		// Create  associated subset
		ConditionArray *cand = candidates->GetNext();
		DataSubset *subset = cand->Apply(mDescDataSet, cand->GetParent()->GetMask());

		if(subset->GetSize() >= mMinCoverage) {
			// Compute quality
			Subgroup *sg = new Subgroup(cand, subset);
			mMeasure->ComputeWeightedQuality(sg, mUseOptimisticEstimate);
			results->Add(sg);

			// Recurse
			if(depth < mMaxDepth && !OutOfTime()) {
				// Use optimistic estimate?
				if(mUseOptimisticEstimate) {
					if(results->GoodEnough( sg->GetOptimisticEstimate() ))
						Expand(depth+1, sg, results); // Only expand if optimistic quality good enough for adding to result collection
				} else { // Do not use optimistic estimate, 
					Expand(depth+1, sg, results); // always expand
				}
			}

			// Cleanup
			delete sg;

		} else {	// discard
			delete cand;
			delete subset;
		}
		++curCand;
	}
	results->AddToCandidateCount(numCands);

	// Cleanup
	delete candidates;
}
