//
// Copyright (c) 2011-2012, Matthijs van Leeuwen
//
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
// Neither the name of the Universiteit Utrecht, KU Leuven, nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
#include "../elobal.h"

// qtils
#include <Config.h>
#include <RandomUtils.h>
#include <StringUtils.h>
#include <logger/Log.h>

// bass
#include <Bass.h>

#include "../datastructs/conditions/ConditionArray.h"
#include "../datastructs/dataset/EMMDataSet.h"
#include "../datastructs/subgroup/SubgroupCollection.h"
#include "../datastructs/subgroup/Subgroup.h"
#include "../measures/QualityMeasure.h"
#include "../models/Model.h"
#include "../postproc/PostProcessor.h"
#include "../search/Search.h"
#include "../search/beam/BeamSelector.h"

#include "DiverseSearchTH.h"

DiverseSearchTH::DiverseSearchTH(Config *conf) : TaskHandler(conf){

}
DiverseSearchTH::~DiverseSearchTH() {
	// not my Config*
}

void DiverseSearchTH::HandleTask() {
	string command = mConfig->Read<string>("command");

			if(command.compare("dssd") == 0)				DSSD();
	else	if(command.compare("evaluateSg") == 0)			EvaluateSubgroups();

	else	throw string(__FUNCTION__) + ": Unable to handle task `" + command + "`";
}

string DiverseSearchTH::BuildWorkingDir() {
	string command = mConfig->Read<string>("command");

	if(command.compare("dssd") == 0 || command.compare("evaluateSg") == 0)
		return  command + "/" + mConfig->Read<string>("dsName") + "-" + Logger::GetStartTimeString() + "/";
	
	return command + "/";
}

void DiverseSearchTH::DSSD() {
	char temp[500];
	string dsName = mConfig->Read<string>("dsName");
	LOG_MSG("\n** Reading dataset ... " + dsName);
	EMMDataSet *emmDS = EMMDataSet::RetrieveDataSet(dsName);

	// Build description and model spaces
	DataSet *descriptionDS = emmDS->CreateDescriptionDataSet();
	DataSet *modelDS = emmDS->CreateModelDataSet();
	ConditionArray::Init(modelDS->GetNumRows());
	
	// -- Phase 1: Perform search
	LOG_MSG("\n** Finished reading data. Starting search.");
	Search *search = Search::Create(mConfig, descriptionDS, modelDS);
	search->LogParameters();
	SubgroupCollection *results = search->DoSearch();
	QualityMeasure *measure = search->GetQualityMeasure();
	LOG_MSG("** Finished search.\n");

	uint32 time1 = search->GetElapsedTime(), time2 = 0, time3 = 0;
	SubgroupCollectionStats stats1 = results->ComputeStats(emmDS->GetNumRows()), stats2, stats3;
	printf_s("Writing subgroups to disk ...\r");
	results->WriteStatsFile(Bass::GetWorkingDir() + "stats1-" + Logger::GetStartTimeString() + ".csv", descriptionDS);
	results->WriteSubgroups(Bass::GetWorkingDir() + "results1-" + Logger::GetStartTimeString() + ".sg", descriptionDS);
	LOG_MSG("Saved subgroups and their stats.");
	
	// -- Phase 2: Post-processing to improve individual results
	bool postProcess = mConfig->KeyExists("postProcess");
	if(postProcess) {
		PostProcessor::Process(mConfig->Read<string>("postProcess"), results, descriptionDS, measure);

		time2 = search->GetElapsedTime();
		stats2 = results->ComputeStats(emmDS->GetNumRows());
		printf_s("Writing subgroups to disk ...\r");
		results->WriteStatsFile(Bass::GetWorkingDir() + "stats2-" + Logger::GetStartTimeString() + ".csv", descriptionDS);
		results->WriteSubgroups(Bass::GetWorkingDir() + "results2-" + Logger::GetStartTimeString() + ".sg", descriptionDS);
		LOG_MSG("Saved subgroups and their stats.");
	}

	// -- Phase 3: Perform subgroup set selection
	uint32 postSelect = mConfig->Read<uint32>("postSelect", 0);
	if(postSelect > 0) {
		LOG_MSG("\n * Performing subgroup selection on current results set using beam selector.");

		// Init
		SubgroupList *subgroups = results->GetSubgroups();
		SubgroupList *selected = new SubgroupList();
		Config *beamConfig = new Config(mConfig);
		if(mConfig->KeyExists("postSelectBeam")) // use default beamType otherwise
			beamConfig->Set("beamType", mConfig->Read<string>("postSelectBeam"));
		beamConfig->Set("beamSize", postSelect);
		BeamSelector *beamSelector = BeamSelector::Create(beamConfig, descriptionDS, modelDS);

		// Do actual selection
		beamSelector->Select(subgroups, selected, mConfig->Read<uint32>("maxDepth"), true);

		// Feed selected subgroups to collection
		subgroups->clear();
		for(SubgroupList::iterator it=selected->begin(); it!=selected->end(); it++)
			subgroups->push_back(*it);
		selected->clear();
		results->UpdateStats();

		// Cleanup
		delete selected;
		delete beamSelector;
		delete beamConfig;

		// Output
		time3 = search->GetElapsedTime();
		stats3 = results->ComputeStats(emmDS->GetNumRows());
		printf_s("Writing subgroups to disk ...\r");
		results->WriteStatsFile(Bass::GetWorkingDir() + "stats3-" + Logger::GetStartTimeString() + ".csv", descriptionDS);
		results->WriteSubgroups(Bass::GetWorkingDir() + "results3-" + Logger::GetStartTimeString() + ".sg", descriptionDS);
		LOG_MSG("Saved subgroups and their stats.");
	}

	// -- Final output
	// Display some results to screen
	LOG_MSG("\n** Search results");
	sprintf_s(temp, 500, "# candidates:\t%u", results->GetNumCandidates()); LOG_MSG(string(temp));
	sprintf_s(temp, 500, "# subgroups:\t%u", results->GetSize()); LOG_MSG(string(temp));
	sprintf_s(temp, 500, "Min quality:\t%.05f", results->GetMinQuality()); LOG_MSG(string(temp));
	sprintf_s(temp, 500, "Max quality:\t%.05f\n", results->GetMaxQuality()); LOG_MSG(string(temp));

	// Write short result file
	string filename = Bass::GetWorkingDir() + "the_numbers.txt";
	FILE *fp;
	fopen_s(&fp, filename.c_str(), "w");
	fprintf_s(fp, "%s\t%s\t%s\t%u\t%u\t%u\t%s\n", mConfig->Read<string>("dsName").c_str(), search->GetName().c_str(), search->GetQualityMeasure()->GetName().c_str(), 
		mConfig->Read<uint32>("minCoverage"), mConfig->Read<uint32>("maxDepth"), mConfig->Read<uint32>("floatNumSplits"), mConfig->Read<string>("maxTime").c_str()
		);
	fprintf_s(fp, "%u\t%u\t%u\t%u\t%u\t%.05f\t%.05f\t%.05f\t%.05f\t%.01f\t%.01f\t%.01f\t%.01f\t%.02f\t%.03f",
		time1, time2, time3, results->GetNumCandidates(), stats1.numSubgroups, stats1.qualityMin, stats1.qualityMax, stats1.qualityAvg, stats1.qualitySd,
		stats1.sizeAvg, stats1.sizeSd, stats1.descAvg, stats1.descSd, stats1.diversityPenalty, stats1.jointEntropy
		);
	if(postProcess) {
		fprintf_s(fp, "\t%u\t%.05f\t%.05f\t%.05f\t%.05f\t%.01f\t%.01f\t%.01f\t%.01f\t%.02f\t%.03f",
			stats2.numSubgroups, stats2.qualityMin, stats2.qualityMax, stats2.qualityAvg, stats2.qualitySd,
			stats2.sizeAvg, stats2.sizeSd, stats2.descAvg, stats2.descSd, stats2.diversityPenalty, stats2.jointEntropy
			);
	}
	if(postSelect > 0) {
		fprintf_s(fp, "\t%u\t%.05f\t%.05f\t%.05f\t%.05f\t%.01f\t%.01f\t%.01f\t%.01f\t%.02f\t%.03f",
			stats3.numSubgroups, stats3.qualityMin, stats3.qualityMax, stats3.qualityAvg, stats3.qualitySd,
			stats3.sizeAvg, stats3.sizeSd, stats3.descAvg, stats3.descSd, stats3.diversityPenalty, stats3.jointEntropy
		);
	}
	fprintf_s(fp, "\n");
	fclose(fp);
	
	// Output rest to disk
	bool outputLabelledSubsets = mConfig->Read<bool>("outputLabelledSubsets", false) && emmDS->HasRowLabels();
	bool outputSubsets = mConfig->Read<bool>("outputSubsets", false);
	if(outputLabelledSubsets || outputSubsets) {
		results->WriteSubsets("subset-" + Logger::GetStartTimeString() + "-", outputSubsets, outputLabelledSubsets, emmDS->GetRowLabels());
		LOG_MSG("Saved subsets belonging to subgroups.");
	}

	if(mConfig->Read<bool>("outputModels", true)) {
		QualityMeasure *measure = search->GetQualityMeasure();
		if(measure->GetModelPolicy() == RebuildModels) {
			printf_s("Rebuilding subgroup models ...\r");
			results->CacheModels(measure);
		}
		results->WriteModels("model-" + Logger::GetStartTimeString() + "-");
		LOG_MSG("Saved models belonging to resulting subgroups.");
		modelDS->GetModel()->WriteHumanReadable(Bass::GetWorkingDir() + "basemodel-" + Logger::GetStartTimeString()); // Let model decide on extension
		LOG_MSG("Saved basemodel (of full dataset).");
	}

	// Cleanup
	ConditionArray::Cleanup();
	delete results;
	delete search;
	delete descriptionDS;
	delete modelDS;
	delete emmDS;
}

void DiverseSearchTH::EvaluateSubgroups() {
	char temp[1000];
	string dsName = mConfig->Read<string>("dsName");
	LOG_MSG("\n** Reading dataset ... " + dsName);
	EMMDataSet *emmDS = EMMDataSet::RetrieveDataSet(dsName);

	// Build description and model spaces
	DataSet *descriptionDS = emmDS->CreateDescriptionDataSet();
	DataSet *modelDS = emmDS->CreateModelDataSet();
	ConditionArray::Init(modelDS->GetNumRows());

	// Create quality measure
	QualityMeasure *measure = QualityMeasure::Create(mConfig, modelDS);

	// Create collection to obtain results
	SubgroupCollection *results = new SubgroupCollection();
	results->SetAutoSort(false);

	// Parse subgroup description from config
	if(mConfig->KeyExists("description") && mConfig->Read<string>("description").size() > 0) {
		LOG_MSG("\n** Evaluating specified subgroup");
		EvaluateSubgroup(results, descriptionDS, measure, mConfig->Read<string>("description"));

	// Read subgroup descriptions from specified sg file
	} else if(mConfig->KeyExists("sgFile") && mConfig->Read<string>("sgFile").size() > 0) {
		LOG_MSG("\n** Evaluating subgroups from specified file");
		FILE *fp; 
		if(fopen_s(&fp, mConfig->Read<string>("sgFile").c_str(), "r") ==0) {
			while(fgets(temp, 1000, fp) != NULL) {
				string line = temp;
				line = StringUtils::Trim(line, " \t\r\n");
				if(line.size() > 2)
					EvaluateSubgroup(results, descriptionDS, measure, line);
			}
			fclose(fp);
		} else
			LOG_MSG("File could not be opened; non-existing?");

	// Read subset from specified file
	} else if(mConfig->KeyExists("subsetFile") && mConfig->Read<string>("subsetFile").size() > 0) {
		LOG_MSG("\n** Evaluating subgroups from specified subset file");
		string filename = mConfig->Read<string>("subsetFile");
		DataSubset *subset = DataSubset::Load(filename, modelDS);
		Subgroup *sg = new Subgroup(new ConditionArray(), subset);

		// Compute quality
		if(subset->GetSize() > 0)
			measure->ComputeQuality(sg);
		else
			sg->SetQuality(std::numeric_limits<double>::quiet_NaN());

		// Add to results
		results->Add(sg); // creates clone
		delete sg;			// so cleanup

	// No input found
	} else {
		LOG_MSG("\n** No subgroup description or sgFile specified.");
	}

	// Only proceed if we have subgroups
	if(results->GetSize() > 0) {
		// Apply post-processing, if asked for
		if(mConfig->KeyExists("postProcess"))
			PostProcessor::Process(mConfig->Read<string>("postProcess"), results, descriptionDS, measure);

		// Computing statistics and scores
		double diversityScore = results->ComputeDiversityScore(emmDS->GetNumRows());

		// Output results
		LOG_MSG("\n** Evaluation results");
		sprintf_s(temp, 1000, "# subgroups:\t%u", results->GetSize()); LOG_MSG(string(temp));
		sprintf_s(temp, 1000, "Min quality:\t%.05f", results->GetMinQuality()); LOG_MSG(string(temp));
		sprintf_s(temp, 1000, "Max quality:\t%.05f", results->GetMaxQuality()); LOG_MSG(string(temp));
		sprintf_s(temp, 1000, "Diversity penalty:\t%.02f", diversityScore); LOG_MSG(string(temp));

		printf_s("Writing subgroups to disk ...\r");
		results->WriteStatsFile(Bass::GetWorkingDir() + "stats-" + Logger::GetStartTimeString() + ".csv", descriptionDS);
		LOG_MSG("Saved subgroups and their stats.");

		results->WriteSubsets("subset-" + Logger::GetStartTimeString() + "-");
		LOG_MSG("Saved subsets belonging to subgroups.");

		results->WriteModels("model-" + Logger::GetStartTimeString() + "-");
		LOG_MSG("Saved models belonging to subgroups.");
	}

	// Cleanup
	ConditionArray::Cleanup();
	delete results;
	delete measure;
	delete descriptionDS;
	delete modelDS;
	delete emmDS;
}
void DiverseSearchTH::EvaluateSubgroup(SubgroupCollection *results, DataSet *descriptionDS, QualityMeasure *measure, string &description) {
	LOG_MSG(" * Description:\t" + description);

	// Create condition array from description
	ConditionArray *conditions = ConditionArray::FromString(description, descriptionDS);
	if(conditions == NULL) {
		LOG_MSG("--> Could not be parsed! <--");
		return;
	}

	// Apply conditions to create subset & subgroup
	DataSubset *subset = conditions->Apply(descriptionDS);
	Subgroup *sg = new Subgroup(conditions, subset);

	// Compute quality
	if(subset->GetSize() > 0)
		measure->ComputeQuality(sg);
	else
		sg->SetQuality(std::numeric_limits<double>::quiet_NaN());

	// Add to results
	results->Add(sg); // creates clone
	delete sg;			// so cleanup
}
